# standard imports
import logging
from datetime import date, datetime
from pathlib import Path
from typing import Any, Dict, Optional

# ogd imports
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class DatasetKey:
    """Simple little class to make logic with dataset keys easier
    """
    def __init__(self, key:str, game_id:str):
    # 1. Get Game ID from key
        if key[:len(game_id)] != game_id:
            Logger.Log(f"Got a mismatch between expected game ID and ID in the dataset key: {key[:len(game_id)]} != {game_id}. Defaulting to {game_id}")
        self._game_id = game_id
    # 2. Get Dates from key
        _date_range = key[len(game_id):]
        _date_range_parts = _date_range.split("_")
        # If this _dataset_key matches the expected format,
        # i.e. spit is: ["", "YYYYMMDD", "to", "YYYYMMDD"]
        if len(_date_range_parts) == 4:
            self._from_year  = int(_date_range_parts[1][0:4])
            self._from_month = int(_date_range_parts[1][4:6])
            self._to_year    = int(_date_range_parts[3][0:4])
            self._to_month   = int(_date_range_parts[3][4:6])
        else:
            self._from_year  = None
            self._from_month = None
            self._to_year    = None
            self._to_month   = None
        self._original_key = key

    def __str__(self):
        return self._original_key
    
    @property
    def IsValid(self) -> bool:
        return  self._from_year  is not None \
            and self._from_month is not None \
            and self._to_year    is not None \
            and self._to_month   is not None
    @property
    def GameID(self) -> str:
        return self._game_id
    @property
    def FromYear(self) -> int:
        return self._from_year or -1
    @property
    def FromMonth(self) -> int:
        return self._from_month or -1
    @property
    def ToYear(self) -> int:
        return self._to_year or -1
    @property
    def ToMonth(self) -> int:
        return self._to_month or -1

class DatasetSchema(Schema):
    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._key                 : DatasetKey
        self._date_modified       : date | str
        self._start_date          : date | str
        self._end_date            : date | str
        self._ogd_revision        : str
        self._session_ct          : Optional[int]
        self._player_ct           : Optional[int]
        self._raw_file            : Optional[Path]
        self._events_file         : Optional[Path]
        self._events_template     : Optional[Path]
        self._sessions_file       : Optional[Path]
        self._sessions_template   : Optional[Path]
        self._players_file        : Optional[Path]
        self._players_template    : Optional[Path]
        self._population_file     : Optional[Path]
        self._population_template : Optional[Path]

        if not isinstance(all_elements, dict):
            all_elements = {}
        _game_id = DatasetSchema._parseGameID(name)
        self._key = DatasetKey(key=name, game_id=_game_id)
    # 1. Parse dates
        if "date_modified" in all_elements.keys():
            self._date_modified = DatasetSchema._parseDateModified(all_elements["date_modified"])
        else:
            self._date_modified = "UNKNOWN"
        if "start_date" in all_elements.keys():
            self._start_date = DatasetSchema._parseStartDate(all_elements["start_date"])
        else:
            self._start_date = "UNKNOWN"
        if "end_date" in all_elements.keys():
            self._end_date = DatasetSchema._parseEndDate(all_elements["end_date"])
        else:
            self._end_date = "UNKNOWN"
    # 2. Parse metadata
        if "ogd_revision" in all_elements.keys():
            self._ogd_revision = DatasetSchema._parseOGDRevision(all_elements["ogd_revision"])
        else:
            self._ogd_revision = "UNKNOWN"
        if "sessions" in all_elements.keys():
            self._session_ct = DatasetSchema._parseSessionCount(all_elements["sessions"])
        else:
            self._session_ct = None
        if "players" in all_elements.keys():
            self._players = DatasetSchema._parsePlayerCount(all_elements["players"])
        else:
            self._player_ct = None
    # 3. Parse file/template paths
        if "raw_file" in all_elements.keys():
            self._raw_file = DatasetSchema._parseRawFile(all_elements["raw_file"])
        else:
            self._raw_file = None
        if "events_file" in all_elements.keys():
            self._events_file = DatasetSchema._parseEventsFile(all_elements["events_file"])
        else:
            self._events_file = None
        if "events_template" in all_elements.keys():
            self._events_template = DatasetSchema._parseEventsTemplate(all_elements["events_template"])
        else:
            self._events_template = None
        if "sessions_file" in all_elements.keys():
            self._sessions_file = DatasetSchema._parseSessionsFile(all_elements["sessions_file"])
        else:
            self._sessions_file = None
        if "sessions_template" in all_elements.keys():
            self._sessions_template = DatasetSchema._parseSessionsTemplate(all_elements["sessions_template"])
        else:
            self._sessions_template = None
        if "players_file" in all_elements.keys():
            self._players_file = DatasetSchema._parsePlayersFile(all_elements["players_file"])
        else:
            self._players_file = None
        if "players_template" in all_elements.keys():
            self._players_template = DatasetSchema._parsePlayersTemplate(all_elements["players_template"])
        else:
            self._players_template = None
        if "population_file" in all_elements.keys():
            self._population_file = DatasetSchema._parsePopulationFile(all_elements["population_file"])
        else:
            self._population_file = None
        if "population_template" in all_elements.keys():
            self._population_template = DatasetSchema._parsePopulationTemplate(all_elements["population_template"])
        else:
            self._population_template = None

        _used = {"date_modified", "start_date", "end_date", "ogd_revision", "sessions", "players",
                 "raw_file", "events_file", "events_template",
                 "sessions_file", "sessions_template", "players_file", "players_template",
                 "population_file", "population_template"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    def __str__(self) -> str:
        return str(self.Key)

    # *** Properties ***

    @property
    def Key(self) -> DatasetKey:
        return self._key
    @property
    def DateModified(self) -> date | str:
        return self._date_modified
    @property
    def DateModifiedStr(self) -> str:
        ret_val : str
        if isinstance(self._date_modified, date):
            ret_val = self._date_modified.strftime("%m/%d/%Y")
        else:
            ret_val = self._date_modified
        return ret_val
    @property
    def StartDate(self) -> date | str:
        return self._start_date
    @property
    def EndDate(self) -> date | str:
        return self._end_date
    @property
    def OGDRevision(self) -> str:
        return self._ogd_revision
    @property
    def SessionCount(self) -> Optional[int]:
        return self._session_ct
    @property
    def PlayerCount(self) -> Optional[int]:
        return self._player_ct
    @property
    def RawFile(self) -> Optional[Path]:
        return self._raw_file
    @property
    def EventsFile(self) -> Optional[Path]:
        return self._events_file
    @property
    def EventsTemplate(self) -> Optional[Path]:
        return self._events_template
    @property
    def SessionsFile(self) -> Optional[Path]:
        return self._sessions_file
    @property
    def SessionsTemplate(self) -> Optional[Path]:
        return self._sessions_template
    @property
    def PlayersFile(self) -> Optional[Path]:
        return self._players_file
    @property
    def PlayersTemplate(self) -> Optional[Path]:
        return self._players_template
    @property
    def PopulationFile(self) -> Optional[Path]:
        return self._population_file
    @property
    def PopulationTemplate(self) -> Optional[Path]:
        return self._population_template

    @property
    def FileSet(self) -> str:
        """
        The list of data files associated with the dataset.

        r -> Raw events file (no generated events)
        e -> All events file (with generated events)
        s -> Session features file
        p -> Player features file
        P -> Popoulation features file

        :return: The list of data files associated with the dataset.
        :rtype: str
        """
        _fset = [
           "r" if self.RawFile is not None else "",
           "e" if self.EventsFile is not None else "",
           "s" if self.SessionsFile is not None else "",
           "p" if self.PlayersFile is not None else "",
           "P" if self.PopulationFile is not None else ""
        ]
        return "".join(_fset)

    @property
    def TemplateSet(self) -> str:
        """
        The list of template files associated with the dataset.

        e -> Events template
        s -> Session features template
        p -> Player features template
        P -> Popoulation features template

        :return: The list of template files associated with the dataset.
        :rtype: str
        """
        _tset = [
           "e" if self.EventsTemplate is not None else "",
           "s" if self.SessionsTemplate is not None else "",
           "p" if self.PlayersTemplate is not None else "",
           "P" if self.PopulationTemplate is not None else ""
        ]
        return "".join(_tset)

    @property
    def AsMarkdown(self) -> str:
        ret_val : str = \
f"""{self.Name}: {self.PlayerCount} players across {self.SessionCount} sessions.  
Last modified {self.DateModified.strftime('%m/%d/%Y') if type(self.DateModified) == date else self.DateModified} with OGD v.{self.OGDRevision or 'UNKNOWN'}  
- Files: [{self.FileSet}]  
- Templates: [{self.TemplateSet}]"""
        return ret_val

    @staticmethod
    def EmptySchema() -> "DatasetSchema":
        return DatasetSchema(name="NOT FOUND", all_elements={})

    # TODO : once we have official minimum Python up to 3.11, import Self and set other:Optional[Self]
    def IsNewerThan(self, other) -> bool | None:
        """
        Check if `self` has a more recent "modified on" date than `other`.

        If `other` is None, returns True by default.  
        If both `self` and `other` are DatasetSchemas, but one (or both) is missing a "modified" date, returns None, because it is indeterminate. 

        :param other: The DatasetSchema to be compared with `self`.
        :type other: Optional[Self]
        :return: True if `self` has a more recent "modified" date than `other`, otherwise False. If one (or both) are missing "modified" date, then None. If `other` is None, True by default.
        :rtype: bool | None
        """
        if other == None:
            return True
        if isinstance(self.DateModified, date) and isinstance(other.DateModified, date):
            return self.DateModified > other.DateModified
        else:
            return None

    # *** Private Functions ***

    # NOTE: Yes, most of these parse functions are redundant, but that's fine,
    # we just want to have one bit of code to parse each piece of the schema, even if most do the same thing.

    @staticmethod
    def _parseGameID(dataset_name) -> str:
        ret_val : str
        if isinstance(dataset_name, str):
            ret_val = dataset_name.split('_')[0]
        else:
            ret_val = str(dataset_name).split('_')[0]
            Logger.Log(f"Dataset name was unexpected type {type(dataset_name)}, defaulting to str(dataset_name).split('_')[0]={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDateModified(date_modified) -> date:
        ret_val : date
        if isinstance(date_modified, date):
            ret_val = date_modified
        else:
            ret_val = datetime.strptime(date_modified, "%m/%d/%Y").date()
            Logger.Log(f"Dataset modified date was unexpected type {type(date_modified)}, defaulting to strptime(date_modified)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseStartDate(start_date) -> date:
        ret_val : date
        if isinstance(start_date, date):
            ret_val = start_date
        else:
            ret_val = datetime.strptime(start_date, "%m/%d/%Y").date()
            Logger.Log(f"Dataset start date was unexpected type {type(start_date)}, defaulting to strptime(start_date)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseEndDate(end_date) -> date:
        ret_val : date
        if isinstance(end_date, date):
            ret_val = end_date
        else:
            ret_val = datetime.strptime(end_date, "%m/%d/%Y").date()
            Logger.Log(f"Dataset end date was unexpected type {type(end_date)}, defaulting to strptime(end_date)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseOGDRevision(revision) -> str:
        ret_val : str
        if isinstance(revision, str):
            ret_val = revision
        else:
            ret_val = str(revision)
            Logger.Log(f"Dataset OGD revision was unexpected type {type(revision)}, defaulting to str(revision)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseSessionCount(sess_ct) -> Optional[int]:
        ret_val : Optional[int]
        if isinstance(sess_ct, int):
            ret_val = sess_ct
        elif sess_ct == None:
            ret_val = None
        else:
            ret_val = int(sess_ct)
            Logger.Log(f"Dataset session count was unexpected type {type(sess_ct)}, defaulting to int(sess_ct)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parsePlayerCount(player_ct) -> Optional[int]:
        ret_val : Optional[int]
        if isinstance(player_ct, int):
            ret_val = player_ct
        elif player_ct == None:
            ret_val = None
        else:
            ret_val = int(player_ct)
            Logger.Log(f"Dataset player count was unexpected type {type(player_ct)}, defaulting to int(player_ct)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseRawFile(raw_file:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if raw_file == None:
            ret_val = None
        elif isinstance(raw_file, Path):
            ret_val = raw_file
        elif isinstance(raw_file, str):
            ret_val = Path(raw_file) if raw_file != "" else None
        else:
            ret_val = Path(str(raw_file))
            Logger.Log(f"Dataset raw event file was unexpected type {type(raw_file)}, defaulting to Path(str(raw_file))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseEventsFile(events_file:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if events_file == None:
            ret_val = None
        elif isinstance(events_file, Path):
            ret_val = events_file
        elif isinstance(events_file, str):
            ret_val = Path(events_file) if events_file != "" else None
        else:
            ret_val = Path(str(events_file))
            Logger.Log(f"Dataset all event file was unexpected type {type(events_file)}, defaulting to Path(str(events_file))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseSessionsFile(sessions_file:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if sessions_file == None:
            ret_val = None
        elif isinstance(sessions_file, Path):
            ret_val = sessions_file
        elif isinstance(sessions_file, str):
            ret_val = Path(sessions_file) if sessions_file != "" else None
        else:
            ret_val = Path(str(sessions_file))
            Logger.Log(f"Dataset session feature file was unexpected type {type(sessions_file)}, defaulting to Path(str(sessions_file))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parsePlayersFile(players_file:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if players_file == None:
            ret_val = None
        elif isinstance(players_file, Path):
            ret_val = players_file
        elif isinstance(players_file, str):
            ret_val = Path(players_file) if players_file != "" else None
        else:
            ret_val = Path(str(players_file))
            Logger.Log(f"Dataset player feature file was unexpected type {type(players_file)}, defaulting to Path(str(players_file))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parsePopulationFile(pop_file:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if pop_file == None:
            ret_val = None
        elif isinstance(pop_file, Path):
            ret_val = pop_file
        elif isinstance(pop_file, str):
            ret_val = Path(pop_file) if pop_file != "" else None
        else:
            ret_val = Path(str(pop_file))
            Logger.Log(f"Dataset population feature file was unexpected type {type(pop_file)}, defaulting to Path(str(pop_file))={ret_val}.", logging.WARN)
        return ret_val


    @staticmethod
    def _parseEventsTemplate(events_tplate:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if events_tplate == None:
            ret_val = None
        elif isinstance(events_tplate, Path):
            ret_val = events_tplate
        elif isinstance(events_tplate, str):
            ret_val = Path(events_tplate) if events_tplate != "" else None
        else:
            ret_val = Path(str(events_tplate))
            Logger.Log(f"Dataset events template was unexpected type {type(events_tplate)}, defaulting to Path(str(events_tplate))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseSessionsTemplate(sessions_tplate:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if sessions_tplate == None:
            ret_val = None
        if isinstance(sessions_tplate, Path):
            ret_val = sessions_tplate
        elif isinstance(sessions_tplate, str):
            ret_val = Path(sessions_tplate) if sessions_tplate != "" else None
        else:
            ret_val = Path(str(sessions_tplate))
            Logger.Log(f"Dataset session template file was unexpected type {type(sessions_tplate)}, defaulting to Path(str(sessions_tplate))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parsePlayersTemplate(players_tplate:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if players_tplate == None:
            ret_val = None
        elif isinstance(players_tplate, Path):
            ret_val = players_tplate
        elif isinstance(players_tplate, str):
            ret_val = Path(players_tplate) if players_tplate != "" else None
        else:
            ret_val = Path(str(players_tplate))
            Logger.Log(f"Dataset players template file was unexpected type {type(players_tplate)}, defaulting to Path(str(players_tplate))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parsePopulationTemplate(pop_tplate:Optional[Path | str]) -> Optional[Path]:
        ret_val : Optional[Path]
        if pop_tplate == None:
            ret_val = None
        elif isinstance(pop_tplate, Path):
            ret_val = pop_tplate
        elif isinstance(pop_tplate, str):
            ret_val = Path(pop_tplate) if pop_tplate != "" else None
        else:
            ret_val = Path(str(pop_tplate))
            Logger.Log(f"Dataset population template file was unexpected type {type(pop_tplate)}, defaulting to Path(str(pop_tplate))={ret_val}.", logging.WARN)
        return ret_val



import os, re, json

def getListOfFiles(dirName):
    # create a list of file and sub directories
    # names in the given directory
    listOfFile = os.listdir(dirName)
    allFiles = list()
    # Iterate over all the entries
    for entry in listOfFile:
        # Create full path
        fullPath = os.path.join(dirName, entry)
        # If entry is a directory then get the list of files in this directory
        if not os.path.isdir(fullPath):
            allFiles.append(fullPath)
        else:
            allFiles += getListOfFiles(fullPath)
    return allFiles

def get_all_items(jowilder_path, outpath):
    # get all files from the scenes folder
    owd = os.getcwd()
    levels_path = os.path.join(jowilder_path, 'assets/data/levels')
    os.chdir(levels_path)
    allfiles = getListOfFiles('tunic/scenes')
    os.chdir(owd)

    # seperate all that to json
    all_items = []
    pattern = re.compile(r"""tmp_.*_command\.raw_[a]*text = "(.*)";\ntmp_.*_command\.speaker = (.*);""")
    for path in allfiles:
        true_path = os.path.join(levels_path, path)
        # get text of any applicable character
        with open(true_path) as f:
            text_speaker = pattern.findall(f.read())

        # make path dictionary
        path_list = path.replace('.meta', '').split('/')
        if len(path_list) > 6:
            all_items.append({
                'path':  true_path,
                'scene': path_list[2],
                'room': path_list[4],
                'type': path_list[5],
                'fqid': '.'.join(path_list[6:]),
                'text_speaker': text_speaker
            })

    with open(outpath, 'w+') as outfile:
        json.dump(all_items, outfile, indent=4)

if __name__ == '__main__':
    jowilder_path = '/Users/johnmccloskey/Development/jo_wilder'
    outpath = 'game_info/Jowilder/jowilder_all_items.json'
    get_all_items(jowilder_path=jowilder_path, outpath=outpath)



# import libraries
import json
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class FinalAttributes(PerLevelFeature):
    def __init__(self, params:GeneratorParameters):
        PerCountFeature.__init__(self, params=params)
        self._ATTRIBUTE_ENUM : List[str] = ["endurance", "resourceful", "tech","social","trust","research"]
        self._last_attribs : Dict[str, Optional[int]] = dict(zip(self._ATTRIBUTE_ENUM, [None]*len(self._ATTRIBUTE_ENUM)))


    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _getEventDependencies(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _getFeatureDependencies(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _extractFromEvent(self, event:Event) -> None:
        #self._story_alignment = event.EventData["story_alignment"]
        _default = str([None]*len(self._ATTRIBUTE_ENUM))
        _stats = json.loads(event.GameState.get("current_stats", _default))
        self._last_attribs = dict(zip(self._ATTRIBUTE_ENUM, _stats))

    def _extractFromFeatureData(self, feature:FeatureData):
        #add logic to make sure that MODE is session, not player so we don't get duplicates
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._last_attribs]

    # *** Optionally override public functions. ***
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.PLAYER, ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<
    



__all__ = [
    "ScenesEncountered",
    "Session_Language",
    "SessionDuration",
    "HeadsetOnCount",
    "SceneFailures",
    "TaskTimeToComplete",
    "SceneDuration",
    "ObjectSelectionsDuringVoiceover",
    "SceneFailureCount"
]
# aggregated features
from . import ScenesEncountered
from . import Session_Language
from . import SessionDuration
from . import HeadsetOnCount
from . import ObjectSelectionsDuringVoiceover
from . import SceneFailureCount
# percount features
from . import SceneFailures
from . import TaskTimeToComplete
from . import SceneDuration



# import libraries
import json
from datetime import datetime
from typing import List, Type, Optional
# import locals
from ogd.core.processors.Processor import Processor
from ogd.core.models.Event import Event
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.utils import ExportRow

class EventProcessor(Processor):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_schema:GameSchema):
        super().__init__(game_schema=game_schema)
        self._events : List[Event] = []

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _processEvent(self, event:Event):
        self._events.append(event)

    def _getLines(self) -> List[ExportRow]:
        return [_event.ColumnValues() for _event in self._events]

    def _clearLines(self) -> None:
        self._events = []

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



__all__ = [
    "Generator",
    "detectors",
    "extractors",
    "registries",
    "GeneratorLoader"
]

from . import Generator
from . import detectors
from . import extractors
from . import registries
from . import GeneratorLoader


# import standard libraries
import logging
from typing import List, Type, Optional, Set
# import local files
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.registries.ExtractorRegistry import ExtractorRegistry
from ogd.core.processors.ExtractorProcessor import ExtractorProcessor
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

## @class PopulationProcessor
#  Class to extract and manage features for a processed csv file.
class PopulationProcessor(ExtractorProcessor):

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the PopulationProcessor class.
    def __init__(self, LoaderClass: Type[GeneratorLoader], game_schema: GameSchema,
                 feature_overrides:Optional[List[str]]=None):
        """Constructor for the PopulationProcessor class.
        Simply stores some data for use later, including the type of extractor to use.

        :param LoaderClass: The type of data extractor to use for input data.
                            This should correspond to whatever game_id is in the TableSchema.
        :type LoaderClass: Type[GeneratorLoader]
        :param game_schema: A dictionary that defines how the game data itself is structured.
        :type game_schema: GameSchema
        :param feature_overrides: _description_, defaults to None
        :type feature_overrides: Optional[List[str]], optional
        :param pop_file: _description_, defaults to None
        :type pop_file: Optional[IO[str]], optional
        """
        self._players  : Set[str] = set()
        self._sessions : Set[str] = set()
        super().__init__(LoaderClass=LoaderClass, game_schema=game_schema, feature_overrides=feature_overrides)

    def __str__(self):
        return f"PopulationProcessor"

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @property
    def _mode(self) -> ExtractionMode:
        return ExtractionMode.POPULATION

    @property
    def _playerID(self) -> str:
        return "population"

    @property
    def _sessionID(self) -> str:
        return "population"

    def _getExtractorNames(self) -> List[str]:
        if isinstance(self._registry, ExtractorRegistry):
            return ["PlayerCount", "SessionCount"] + self._registry.GetExtractorNames()
        else:
            raise TypeError("PopulationProcessor's registry is not a ExtractorRegistry!")

    ## Function to handle processing of a single row of data.
    def _processEvent(self, event:Event):
        """Function to handle processing of a single row of data.
        Basically just responsible for ensuring an extractor for the session corresponding
        to the row already exists, then delegating the processing to that extractor.

        :param event: An object with the data for the event to be processed.
        :type event: Event
        """
        if event.UserID:
            self._players.add(event.UserID)
        self._sessions.add(event.SessionID)
        self._registry.UpdateFromEvent(event=event)

    def _getLines(self) -> List[ExportRow]:
        ret_val : ExportRow
        # if as_str:
        #     ret_val = [str(len(self._players)), str(len(self._sessions))] + self._registry.GetFeatureStringValues()
        # else:
        ret_val = [len(self._players), len(self._sessions)] + self._registry.GetFeatureValues()
        return [ret_val]

    def _getFeatureData(self, order:int) -> List[FeatureData]:
        return self._registry.GetFeatureData(order=order)

    ##  Function to empty the list of lines stored by the PopulationProcessor.
    #   This is helpful if we're processing a lot of data and want to avoid
    #   eating too much memory.
    def _clearLines(self) -> None:
        Logger.Log(f"Clearing features from PopulationProcessor.", logging.DEBUG, depth=2)
        self._registry = ExtractorRegistry(mode=self._mode)

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***


from typing import Any, List, Optional

from ogd.core.models.enums.ExtractionMode import ExtractionMode

class FeatureData:
   def __init__(self, name:str, feature_type:str, count_index:Optional[int],
                cols:List[str], vals:List[Any],   mode:ExtractionMode,
                player_id:Optional[str]=None,     sess_id:Optional[str]=None):
      self._name = name
      self._feature_type = feature_type
      self._count_index = count_index
      self._cols = cols
      self._vals = vals
      self._mode = mode
      self._player_id = player_id
      self._sess_id = sess_id

   def __str__(self):
      return f"Name: {self.Name}\tCount Index: {self.CountIndex}\nColumns: {self._cols}\t Values: {self._vals}\nMode: {self._mode.name}\tPlayer: {self.PlayerID}\tSession: {self.SessionID}"

   def __repr__(self):
      return self.Name

   @property
   def Name(self):
      return self._name

   @property
   def FeatureType(self):
      return self._feature_type

   @property
   def CountIndex(self):
      return self._count_index

   @property
   def FeatureNames(self) -> List[str]:
      return self._cols

   @property
   def FeatureValues(self) -> List[Any]:
      return self._vals

   @property
   def ExportMode(self):
      return self._mode

   @property
   def PlayerID(self) -> Optional[str]:
      return self._player_id

   @property
   def SessionID(self) -> Optional[str]:
      return self._sess_id


# import libraries
import inspect
import logging
from datetime import timedelta
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.Extractor import GeneratorParameters
from ogd.games.AQUALAB.features.PerJobFeature import PerJobFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature

class RegionJobCount(PerCountFeature):
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        regions = ['arctic', 'coral', 'bayou', 'kelp', 'other']
        self.count = 0
        self.region = regions[self.CountIndex]
    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["complete_job"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _validateEventCountIndex(self, event:Event):
        if event.app_version == 'Aqualab' or event.app_version == 'None':
            if event.EventData.get('job_name', {}).get('string_value').startswith(self.region):
                return True
            else:
                return False
        else:
            if event.EventData.get('job_name', {}).startswith(self.region):
                return True
            else:
                return False
# 'aqualab' and 'GameState"
    def _updateFromEvent(self, event:Event) -> None:
        self.count += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self.count]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.SessionFeature import SessionFeature

# gaze_dict = { 'BigRock00': 0, 'Bridge': 0, 'Bridge2': 0, 'BigRock01': 0, 'River3': 0, 'Inland2': 0, 'Sea4': 0, 'River2': 0,'Inland1': 0, 'Sea5': 0, 'Sea3': 0, 'River5': 0, 'River4': 0, 'Inland5': 0, 'Inland4': 0, 'Inland3': 0,'River1': 0, 'Sea2': 0, 'Sea1': 0}
class GazeCount(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        # self._current_count : int = 0
        self._object_id = None
        self._gaze_dict = dict()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["gaze_object_end"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        # self._current_count += 1
        
        self._object_id = event.event_data.get("object_id")
        if not self._object_id in self._gaze_dict.keys():
                self._gaze_dict[self._object_id] = 0
        
        self._gaze_dict[self._object_id]+=1
        
    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._gaze_dict]



from typing import List, Optional, Dict, Any
import pandas as pd
from datetime import datetime as dt
import json
from models.SequenceModel import SequenceModel

## @class ActionsLastXSecondsModel
# Returns a dictionary of active events for the last X seconds
# @param levels: Levels applicable for model, time: last x seconds
class ActionsLastXSecondsModel(SequenceModel):

    def __init__(self, time=30, levels: List[int] = []):
        '''
        @class MoneyAccumulationModel
        Returns a tuple with (1) the current amount of money and (2) the total number of homes,
        farms, and dairy farms
        :param levels: Levels applicable for model
        '''
        super().__init__()
        self.time = time
        self._ACTIVE_EVENTS = {3: "SELECTTILE", 4: "SELECTFARMBIT", 5: "SELECTITEM", 6: "SELECTBUY", 7: "BUY",
                          8: "CANCELBUY", 10: "TILEUSESELECT", 11: "ITEMUSESELECT", 1: 'STARTGAME',
                          12: 'TOGGLENUTRITION', 13: "TOGGLESHOP", 14: "TOGGLEACHIEVEMENTS",
                          15: "SKIPTUTORIAL", 16: "SPEED"}

    def get_utc(self, data):
        dt_string = '%Y-%m-%d %H:%M:%S'
        return dt.strptime(data, dt_string).timestamp()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False):
        assert events
        return self.get_res(events)

    def get_res(self, events):
        res_dict = {}
        now = dt.utcnow().timestamp()
        for row in events:
            evt_time = self.get_utc(row['client_time'])
            if (evt_time > (now-self.time)) and (evt_time < now):
                if row['event_custom'] in self._ACTIVE_EVENTS.keys():
                    evt = self._ACTIVE_EVENTS[row['event_custom']]
                    event_val = f'{evt}_cnt'
                    if event_val not in res_dict:
                        res_dict[event_val] = 1
                    else:
                        res_dict[event_val] += 1
        res_dict['actions_per_secs'] = sum(res_dict.values()) / self.time
        return res_dict




## import standard libraries
import logging
import math
import re
import typing
## import local libraries
from models.FeatureModel import FeatureModel

## @class LogisticModel
class LinearModel(FeatureModel):
    def __init__(self, coefficient_map: typing.Dict[str, float], levels: typing.List[int] = []):
        super().__init__(levels)
        self._coeff_map = coefficient_map

    ## Function to evaluate a logistic regression model, creating a prediction.
    #  This is based around the equation for probability of Y=1, denoted as p:
    #  p = 1 / (1 + e^-logit(X)),
    #  where X is the input data used to predict Y, and
    #  logit(X) = b0 + b1*x1 + b2*x2 + ... + bn*xn,
    #  where bi are the coefficients.
    #  Based on information at https://www.medcalc.org/manual/logistic_regression.php
    def _eval(self, row: typing.Dict):
        ret_val = 0
        for coeff_name in self._coeff_map.keys():
            # case where coefficient is a normal feature
            if coeff_name in row.keys():
                try:
                    ret_val += self._coeff_map[coeff_name] * row[coeff_name]
                except Exception as err:
                    print(f"Got {type(err)} error when trying to add {coeff_name} term. Value is {row[coeff_name]}. Type is {type(row[coeff_name])}")
                    raise err
            # enum case, where we have coefficient = feature_name.enum_val
            # e.g. maybe we've got a feature that returns a classification as either Foo or Bar, and we want Foo's to count as 0.5 towards the model.
            # then we'll have coefficient 0.5 = classification.Foo.
            # in the code here, we get logit += 0.5 * 1.0 for Foo, and 0.5 * 0.0 for Bar.
            # I'm mildly confused as to why I put this case in here, must have found some case where it could occur?
            elif re.search("\w+\.\w+", coeff_name):
                pieces = coeff_name.split(".")
                if pieces[0] in row.keys():
                    ret_val += self._coeff_map[coeff_name] * (1.0 if row[pieces[0]] == pieces[1] else 0.0)
                else:
                    print(f"Found an element of model that is not a feature: {coeff_name}")
            # specific cases, where we just hardcode a thing that must be consistent across all models.
            elif coeff_name == "Intercept":
                ret_val += self._coeff_map[coeff_name]
            elif coeff_name == "display_name":
                pass
            # default case, print a line.
            else:
                print(f"Found an element of model that is not a feature: {coeff_name}")
        return ret_val


## import standard libraries
import abc
import typing
from typing import List, Dict, Any
import logging
## import local libraries
from models.Model import *

## @class FeatureModel
#  Abstract base class for models that use per-session feature data as input rows.
#  For these models, we wrap the functionality in a private _eval function, which is called once per row.
#
class SequenceModel(Model):
    def __init__(self, levels: typing.List[int] = []):
        super().__init__(levels=levels, input_type=ModelInputType.SEQUENCE)

    def Eval(self, rows: typing.List) -> typing.List:
        return self._eval(rows)

    ## Abstract declaration of a function to perform calculation of a model results from a row.
    #
    #  @param rows : A list of rows of data for a session, which should each be a mapping of column names to values.
    #  row keys include:
    #  id, app_id, app_id_fast, app_version, sess_id, persistent_sess_id, player_id, level, event,
    #  event_custom, event_data_simple, event_data_complex, client_time, client_secs_ms, server_time, remote_addr,
    #  req_id, sess_n, http_user_agent
    #  @return     : A result for the given row of data
    @abc.abstractmethod
    def _eval(self, rows: List[Dict[str, Any]]) -> Any:
        pass


__all__ = [
    "Processor",
    "SessionProcessor",
    "PlayerProcessor",
    "PopulationProcessor"
]

from . import Processor
from . import SessionProcessor
from . import PlayerProcessor
from . import PopulationProcessor


import logging
from typing import Any, List

from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from utils.Logger import Logger



class InSchoolSessions(SessionFeature):
    """A feature that determines whether a session started in school or not."""

import pandas as pd
import pytz
import json
import string
from geopy.geocoders import Nominatim
from datetime import datetime
from timezonefinder import TimezoneFinder

##############################

#FOR LOCALTIME TRIAL with colab code trying:

class PlayLocations(SessionFeature):
    # a session started in school or not.

      # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _getEventDependencies(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _getFeatureDependencies(cls, mode:ExtractionMode) -> List[str]:
        return []
    
    def _extractFromFeatureData(self, feature:FeatureData):
        return

    def _extractFromEvent(self, event:Event) -> None:
        if not event.SessionID in self._seen_sessions:
            self._seen_sessions.add(event.SessionID)
            # Step 1: calculate local time
            lat_long = self.calculate_coordinates(event=event)
            local_time = PlayLocations.calculate_local_time_by_coordinates(
            utc_time=event.Timestamp, latitude=lat_long.get('latitude'), longitude=lat_long.get('longitude'))
            # Step 2: check if local time was a school time or not, and add to lists
            self._session_times.append(local_time)
            # Step 3: Check if local time is on a weekday and between 9 AM and 3 PM
            is_weekday = local_time.weekday() < 5  # Monday to Friday is 0 to 4
            is_school_hours = 9 <= local_time.hour < 15
                
            in_school = is_weekday and is_school_hours
            self._in_school_sessions.append(in_school)

            
    def _getFeatureValues(self) -> List[Any]:
        # Sessions that started in school
        return [self._in_school_sessions, self._session_times]

    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)

        # sessions that have been seen before.
        self._seen_sessions = set()
        # sessions that started in school.
        self._in_school_sessions = []
        self._session_times = []
        # geolocater
        self.geolocator = Nominatim(user_agent="timezone_converter")

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["LocalTime"]

    @classmethod
    def AvailableModes(cls) -> List[ExtractionMode]:
        return [ExtractionMode.PLAYER]

    # *** private helper functions ***

    def calculate_coordinates(self, event:Event):
        region = event.UserData.get("region")
        
        # You can continue with the geolocation code
        location_geocoded = self.geolocator.geocode(region, timeout=10) if region else None

        # Latitude and longitude from the geocoded
        latitude  = location_geocoded.latitude  if location_geocoded else None
        longitude = location_geocoded.longitude if location_geocoded else None

        # Filter rows where latitude is not NaN
        return {'latitude':latitude, 'longitude':longitude}


    @staticmethod
    def calculate_local_time_by_coordinates(utc_time, latitude, longitude):
        if latitude is None or longitude is None:
            return None    
        
        tz_finder = TimezoneFinder()
        timezone_str = tz_finder.timezone_at(lng=longitude, lat=latitude)

        if timezone_str:
            local_timezone = pytz.timezone(timezone_str)
            localized_time = pytz.timezone('UTC').localize(utc_time)
            local_time     = localized_time.astimezone(local_timezone)
            return local_time
        else:
            Logger.Log(f"Could not get timezone string, defaulting to datetime.min with UTC timezone", level=logging.WARN)
            return datetime.min.replace(tzinfo=pytz.UTC)  # Default value 







# import standard libraries
import logging
from pathlib import Path
from typing import Any, Dict, Optional, Type
# import local files
from ogd.core.schemas.configs.IndexingSchema import FileIndexingSchema
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.schemas.configs.data_sources.DataSourceSchema import DataSourceSchema
from ogd.core.schemas.configs.data_sources.BigQuerySourceSchema import BigQuerySchema
from ogd.core.schemas.configs.data_sources.FileSourceSchema import FileSourceSchema
from ogd.core.schemas.configs.data_sources.MySQLSourceSchema import MySQLSchema
from ogd.core.schemas.configs.LegacyConfigSchema import LegacyConfigSchema
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class ConfigSchema(Schema):
    """Dumb struct containing properties for each standard OGD-core config item.
    """
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        """Constructs a ConfigSchema, from a name and dictionary of JSON-style elements.

        :param name: The name of the configuration schema.
        :type name: str
        :param all_elements: A dictionary mapping config item names to their configured values. These will be checked and made available through the ConfigSchema properties.
        :type all_elements: Dict[str, Any]
        """
        self._log_file   : bool
        self._batch_size : int
        self._dbg_level  : int
        self._fail_fast  : bool
        self._file_idx   : FileIndexingSchema
        self._data_src   : Dict[str, DataSourceSchema]
        self._game_src_map : Dict[str, GameSourceSchema]

        self._legacy_elems : LegacyConfigSchema = LegacyConfigSchema(name=f"{name} Legacy", all_elements=all_elements)
        if "LOG_FILE" in all_elements.keys():
            self._log_file = ConfigSchema._parseLogFile(all_elements["LOG_FILE"])
        else:
            self._log_file = False
            Logger.Log(f"{name} config does not have a 'LOG_FILE' element; defaulting to log_file={self._log_file}", logging.WARN)
        if "BATCH_SIZE" in all_elements.keys():
            self._batch_size = ConfigSchema._parseBatchSize(all_elements["BATCH_SIZE"])
        else:
            self._batch_size = 500
            Logger.Log(f"{name} config does not have a 'BATCH_SIZE' element; defaulting to batch_size={self._batch_size}", logging.WARN)
        if "DEBUG_LEVEL" in all_elements.keys():
            self._dbg_level = ConfigSchema._parseDebugLevel(all_elements["DEBUG_LEVEL"])
        else:
            self._dbg_level = logging.INFO
            Logger.Log(f"{name} config does not have a 'DEBUG_LEVEL' element; defaulting to dbg_level={self._dbg_level}", logging.WARN)
        if "FAIL_FAST" in all_elements.keys():
            self._fail_fast = ConfigSchema._parseFailFast(all_elements["FAIL_FAST"])
        else:
            self._fail_fast = False
            Logger.Log(f"{name} config does not have a 'FAIL_FAST' element; defaulting to fail_fast={self._fail_fast}", logging.WARN)
        if "FILE_INDEXING" in all_elements.keys():
            self._file_idx = ConfigSchema._parseFileIndexing(all_elements["FILE_INDEXING"])
        else:
            _fallback_elems = { "LOCAL_DIR" : self._legacy_elems.DataDirectory }
            self._file_idx = FileIndexingSchema(name="FILE_INDEXING", all_elements=_fallback_elems)
            Logger.Log(f"{name} config does not have a 'FILE_INDEXING' element; defaulting to file_indexing={self._file_idx}", logging.WARN)
        if "GAME_SOURCES" in all_elements.keys():
            self._data_src = ConfigSchema._parseDataSources(all_elements["GAME_SOURCES"])
        else:
            self._data_src = {}
            Logger.Log(f"{name} config does not have a 'GAME_SOURCES' element; defaulting to game_sources={self._data_src}", logging.WARN)
        if "GAME_SOURCE_MAP" in all_elements.keys():
            self._game_src_map = ConfigSchema._parseGameSourceMap(map=all_elements["GAME_SOURCE_MAP"], sources=self._data_src)
        else:
            self._game_src_map = {}
            Logger.Log(f"{name} config does not have a 'GAME_SOURCE_MAP' element; defaulting to game_source_map={self._game_src_map}", logging.WARN)

        _used = {"LOG_FILE", "BATCH_SIZE", "DEBUG_LEVEL", "FAIL_FAST", "FILE_INDEXING", "GAME_SOURCES", "GAME_SOURCE_MAP"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def DataDirectory(self) -> Path:
        """
        The local directory where export files will be stored by default.

        :return: The local directory where export files will be stored by default.
        :rtype: Path
        """
        return self.FileIndexConfig.LocalDirectory

    @property
    def UseLogFile(self) -> bool:
        """
        Whether to use a log file to store log output from exports.
        """
        return self._log_file

    @property
    def BatchSize(self) -> int:
        """
        The number of sessions to process at once during an export.
        """
        return self._batch_size

    @property
    def DebugLevel(self) -> int:
        """
        The least-severe level of debug output that should be logged.
        Defaults to INFO if not specified.
        """
        return self._dbg_level

    @property
    def FailFast(self) -> bool:
        """
        Whether to fail the export on errors due to bad data, or ignore the bad data and continue processing.
        """
        return self._fail_fast

    @property
    def FileIndexConfig(self) -> FileIndexingSchema:
        """
        A collection of settings for indexing output files.

        TODO : Need better documentation of this item.
        """
        return self._file_idx

    @property
    def DataSources(self) -> Dict[str, DataSourceSchema]:
        """
        A collection of all configured sources of data that can be used for exports.
        """
        return self._data_src

    @property
    def GameSourceMap(self) -> Dict[str, GameSourceSchema]:
        """
        A mapping from game IDs to the data sources they use.
        """
        return self._game_src_map

    @property
    def AsMarkdown(self) -> str:
        """
        A Markdown-formatted stringification of the ConfigSchema.
        (presently just the schema name)
        """
        ret_val : str

        ret_val = f"{self.Name}"
        return ret_val

    @staticmethod
    def _parseLogFile(use_log) -> bool:
        ret_val : bool
        if isinstance(use_log, bool):
            ret_val = use_log
        else:
            ret_val = False
            Logger.Log(f"Config to use log file was unexpected type {type(use_log)}, defaulting to False.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseBatchSize(batch_size) -> int:
        ret_val : int
        if isinstance(batch_size, int):
            ret_val = batch_size
        elif isinstance(batch_size, str):
            ret_val = int(batch_size)
        else:
            ret_val = int(str(batch_size))
            Logger.Log(f"Config batch size was unexpected type {type(batch_size)}, defaulting to int(str(batch_size))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDebugLevel(level) -> int:
        ret_val : int
        if isinstance(level, str):
            match level.upper():
                case "ERROR":
                    ret_val = logging.ERROR
                case "WARNING" | "WARN":
                    ret_val = logging.WARN
                case "INFO":
                    ret_val = logging.INFO
                case "DEBUG":
                    ret_val = logging.DEBUG
                case _:
                    ret_val = logging.INFO
                    Logger.Log(f"Config debug level had unexpected value {level}, defaulting to logging.INFO.", logging.WARN)
        else:
            ret_val = logging.INFO
            Logger.Log(f"Config debug level was unexpected type {type(level)}, defaulting to logging.INFO.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseFailFast(fail_fast) -> bool:
        ret_val : bool
        if isinstance(fail_fast, bool):
            ret_val = fail_fast
        elif isinstance(fail_fast, str):
            match fail_fast.upper():
                case "TRUE":
                    ret_val = True
                case "FALSE":
                    ret_val = False
                case _:
                    ret_val = False
                    Logger.Log(f"Config fail fast had unexpected value {fail_fast}, defaulting to False.", logging.WARN)
        else:
            ret_val = False
            Logger.Log(f"Config fail fast was unexpected type {type(fail_fast)}, defaulting to False.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseFileIndexing(indexing) -> FileIndexingSchema:
        ret_val : FileIndexingSchema
        if isinstance(indexing, dict):
            ret_val = FileIndexingSchema(name="FILE_INDEXING", all_elements=indexing)
        else:
            ret_val = FileIndexingSchema(name="FILE_INDEXING", all_elements={})
            Logger.Log(f"Config file indexing was unexpected type {type(indexing)}, defaulting to default indexing config: {ret_val.AsMarkdown}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDataSources(sources) -> Dict[str, DataSourceSchema]:
        ret_val : Dict[str, DataSourceSchema]
        if isinstance(sources, dict):
            ret_val = {}
            for key,val in sources.items():
                match (val.get("DB_TYPE", "").upper()):
                    case "BIGQUERY" | "FIREBASE":
                        ret_val[key] = BigQuerySchema(name=key, all_elements=val)
                    case "MYSQL":
                        ret_val[key] = MySQLSchema(name=key, all_elements=val)
                    case "FILE":
                        ret_val[key] = FileSourceSchema(name=key, all_elements=val)
                    case _:
                        Logger.Log(f"Game source {key} did not  have a valid 'DB_TYPE' (value: {val.get('DB_TYPE', '')}), and will be skipped!", logging.WARN)
        else:
            ret_val = {}
            Logger.Log(f"Config data sources was unexpected type {type(sources)}, defaulting to empty dict: {ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseGameSourceMap(map, sources) -> Dict[str, GameSourceSchema]:
        ret_val : Dict[str, GameSourceSchema]
        if isinstance(map, dict):
            ret_val = { key : GameSourceSchema(name=key, all_elements=val, data_sources=sources) for key, val in map.items() }
        else:
            ret_val = {}
            Logger.Log(f"Config game source map was unexpected type {type(map)}, defaulting to empty dict: {ret_val}.", logging.WARN)
        return ret_val


from datetime import timedelta
from typing import Any, List

from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class UserTotalSessionDuration(SessionFeature):
    """_summary_

    :param SessionFeature: _description_
    :type SessionFeature: _type_
    """
    def __init__(self, params:GeneratorParameters, player_id:str):
        super().__init__(params=params)
        self._player_id = player_id
        self._time = timedelta(0)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["SessionDuration"]

    def _updateFromEvent(self, event:Event) -> None:
        return

    def _updateFromFeatureData(self, feature:FeatureData):
        if feature.PlayerID == self._player_id:
            if type(feature.FeatureValues[0]) == str and feature.FeatureValues[0] == "No events":
                pass
            else:
                self._time += feature.FeatureValues[0]

    def _getFeatureValues(self) -> List[Any]:
        return [self._time.total_seconds()]

    # *** Optionally override public functions. ***



# import standard libraries
import abc
from datetime import datetime
from typing import Dict, List, Optional, Set
# import local files
from ogd.core.interfaces.DataInterface import DataInterface
from ogd.core.interfaces.outerfaces.DataOuterface import DataOuterface
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.utils.Logger import Logger

class ExporterRange:
    """
    Simple class to define a range of data for export.
    """
    def __init__(self, date_min:Optional[datetime], date_max:Optional[datetime], ids:Optional[List[str]], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]]=None):
        self._date_min : Optional[datetime] = date_min
        self._date_max : Optional[datetime] = date_max
        self._ids      : Optional[List[str]] = ids
        self._id_mode  : IDMode                = id_mode
        self._versions : Optional[List[int]] = versions

    @staticmethod
    def FromDateRange(source:DataInterface, date_min:datetime, date_max:datetime, versions:Optional[List[int]]=None):
        ids = source.IDsFromDates(date_min, date_max, versions=versions)
        return ExporterRange(date_min=date_min, date_max=date_max, ids=ids, id_mode=IDMode.SESSION, versions=versions)

    @staticmethod
    def FromIDs(source:DataInterface, ids:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]]=None):
        date_range = source.DatesFromIDs(id_list=ids, id_mode=id_mode, versions=versions)
        return ExporterRange(date_min=date_range['min'], date_max=date_range['max'], ids=ids, id_mode=id_mode, versions=versions)

    @property
    def DateRange(self) -> Dict[str,Optional[datetime]]:
        return {'min':self._date_min, 'max':self._date_max}

    @property
    def IDs(self) -> Optional[List[str]]:
        return self._ids

    @property
    def IDMode(self):
        return self._id_mode

## @class Request
#  Dumb struct to hold data related to requests for data export.
#  This way, we've at least got a list of what is available in a request.
#  Acts as a base class for more specific types of request.
class Request(abc.ABC):
    ## Constructor for the request base class.
    #  Just stores whatever data is given. No checking done to ensure we have all
    #  necessary data, this can be checked wherever Requests are actually used.
    #  @param game_id An identifier for the game from which we want to extract data.
    #                 Should correspond to the app_id in the database.
    #  @param start_date   The starting date for our range of data to process.
    #  @param end_date     The ending date for our range of data to process.
    def __init__(self, range:ExporterRange, exporter_modes:Set[ExportMode],
                interface:DataInterface,    outerfaces:Set[DataOuterface],
                feature_overrides:Optional[List[str]]=None):
        # TODO: kind of a hack to just get id from interface, figure out later how this should be handled.
        self._game_id        : str                    = str(interface._game_id)
        self._interface      : DataInterface          = interface
        self._range          : ExporterRange          = range
        self._exports        : Set[ExportMode]        = exporter_modes
        self._outerfaces     : Set[DataOuterface]     = outerfaces
        self._feat_overrides : Optional[List[str]]    = feature_overrides

    ## String representation of a request. Just gives game id, and date range.
    def __str__(self):
        _fmt = "%Y-%m-%d"
        _min = "Unkown"
        _max = "Unkown"
        try:
            _range = self.Range.DateRange
            _min = _range['min'].strftime(_fmt) if _range['min'] is not None else "None"
            _max = _range['max'].strftime(_fmt) if _range['max'] is not None else "None"
        except Exception as err:
            Logger.Log(f"Got an error when trying to stringify a Request: {type(err)} {str(err)}")
        finally:
            return f"{self._game_id}: {_min}<->{_max} ({[str(export) for export in self._exports]})"

    @property
    def GameID(self):
        return self._game_id

    @property
    def Interface(self) -> DataInterface:
        return self._interface

    @property
    def Range(self) -> ExporterRange:
        return self._range

    @property
    def ExportRawEvents(self) -> bool:
        return ExportMode.EVENTS in self._exports
    @property
    def ExportProcessedEvents(self) -> bool:
        return ExportMode.DETECTORS in self._exports
    @property
    def ExportSessions(self) -> bool:
        return ExportMode.SESSION in self._exports
    @property
    def ExportPlayers(self) -> bool:
        return ExportMode.PLAYER in self._exports
    @property
    def ExportPopulation(self) -> bool:
        return ExportMode.POPULATION in self._exports

    @property
    def Outerfaces(self) -> Set[DataOuterface]:
        return self._outerfaces

    def RemoveExportMode(self, mode:ExportMode):
        self._exports.discard(mode)
        for outerface in self.Outerfaces:
            outerface.RemoveExportMode(mode=mode)

    ## Method to retrieve the list of IDs for all sessions covered by the request.
    #  Note, this will use the 
    def RetrieveIDs(self) -> Optional[List[str]]:
        return self.Range.IDs


# import libraries
import logging
from datetime import datetime, timedelta
from statistics import stdev
from typing import Any, List, Optional
from ogd.core.models.enums.ExtractionMode import ExtractionMode
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobsAttempted(Feature):

    def __init__(self, params:GeneratorParameters, job_map:dict, diff_map: dict):
        self._player_id = None

        self._job_map = job_map
        self._user_code = None
        self._session_id = None
        super().__init__(params=params)

        # Subfeatures
        if self.CountIndex is not None:
            self._job_id = self.CountIndex
            self._job_name = list(job_map.keys())[self.CountIndex]
            self._difficulties = diff_map[self.CountIndex]
        else:
            raise ValueError("JobsAttempted was not given a count index!")
        self._num_starts = 0
        self._num_completes = 0
        self._percent_complete = 0
        self._avg_time_complete = 0
        self._std_dev_complete = 0

        # Time
        self._times : List[int] = []
        # self._time = 0
        # self._job_start_time : Optional[datetime] = None
        # self._prev_timestamp : Optional[datetime] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["accept_job", "complete_job"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["JobActiveTime"]

    def _updateFromEvent(self, event:Event) -> None:
        if event.UserID != self._player_id:
            self._player_id = event.UserID
        if event.SessionID != self._session_id:
            self._session_id = event.SessionID

            # if self._job_start_time is not None and self._prev_timestamp is not None:
                # self._time += (self._prev_timestamp - self._job_start_time).total_seconds()
                # self._job_start_time = event.Timestamp

        _current_job = event.GameState.get('job_name', event.EventData.get('job_name', None))
        if _current_job is None:
            raise KeyError("Could not find key 'job_name' in GameState or EventData!")
        if self._validate_job(_current_job):
            user_code = event.UserID
            job_id = self._job_map[_current_job]

            if event.EventName == "accept_job" and job_id == self._job_id:
                self._num_starts += 1
                self._user_code = user_code
                # self._job_start_time = event.Timestamp

            elif event.EventName == "complete_job" and job_id == self._job_id and user_code == self._user_code:
                self._num_completes += 1

                # if self._job_start_time:
                    # self._time += (event.Timestamp - self._job_start_time).total_seconds()
                    # self._times.append(self._time)
                    # self._time = 0
                    # self._job_start_time = None

        # self._prev_timestamp = event.Timestamp

    def _updateFromFeatureData(self, feature:FeatureData):
        if feature.FeatureType == "JobActiveTime":
            if feature.CountIndex == self.CountIndex:
                _active_time = feature.FeatureValues[0]
                if self.ExtractionMode == ExtractionMode.SESSION \
                and feature.ExportMode == ExtractionMode.SESSION:
                    # session should only have one time, namely the time for the session.
                    self._times = [_active_time]
                    # print(f"JobsAttempted got session-session for player {self._player_id}")
                elif self.ExtractionMode == ExtractionMode.PLAYER \
                and feature.ExportMode   == ExtractionMode.PLAYER:
                    # player should only have one time, namely the time for the player.
                    self._times = [_active_time]
                    # print(f"JobsAttempted got player-player for player {self._player_id}")
                elif self.ExtractionMode == ExtractionMode.POPULATION \
                and feature.ExportMode   == ExtractionMode.PLAYER:
                    # population could have many times. Only add to list if they actually spent time there, though.
                    if _active_time > 0:
                        self._times.append(_active_time)
                    # print(f"JobsAttempted got population-player for player {self._player_id}")
                # else:
                    # print(f"JobsAttempted got a {self.ExtractionMode.name}-{feature.ExportMode.name} matching, not helpful.")
        else:
            print(f"JobsAttempted got a feature of wrong type: {feature.FeatureType}")

    def _getFeatureValues(self) -> List[Any]:
        if self._num_starts > 0:
            self._percent_complete = (self._num_completes / self._num_starts) * 100

        # if self._time != 0:
        #     self._times.append(self._time)
        
        if len(self._times) > 0:
            self._avg_time_complete = sum(self._times) / len(self._times)

        if len(self._times) > 1:
            self._std_dev_complete = stdev(self._times)

        return [self._job_id, self._job_name, self._num_starts, self._num_completes, self._percent_complete, self._avg_time_complete, self._std_dev_complete, self._difficulties]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["job-name", "num-starts", "num-completes", "percent-complete", "avg-time-per-attempt", "std-dev-per-attempt", "job-difficulties"]

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"

    # *** Other local functions
    def _validate_job(self, job_data):
        ret_val : bool = False
        if job_data and job_data in self._job_map:
            ret_val = True
        else:
            self.WarningMessage(f"Got invalid job_name data in JobsAttempted")
        return ret_val



# import standard libraries
import logging
from typing import Any, Dict, List, Optional
# import local files
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class DataElementSchema(Schema):
    """
    Dumb struct to contain a specification of a data element from the EventData, GameState, or UserData attributes of an Event.
    """
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._type        : str
        self._description : str
        self._details     : Optional[Dict[str, str]]

        if not isinstance(all_elements, dict):
            if isinstance(all_elements, str):
                all_elements = { 'description' : all_elements }
                Logger.Log(f"For EventDataElement config of `{name}`, all_elements was a str, probably in legacy format. Defaulting to all_elements = {'{'} description : {all_elements['description']} {'}'}", logging.WARN)
            else:
                all_elements = {}
                Logger.Log(f"For EventDataElement config of `{name}`, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "type" in all_elements.keys():
            self._type = DataElementSchema._parseElementType(all_elements['type'])
        else:
            self._type = "Unknown"
            Logger.Log(f"{name} EventDataElement config does not have a 'type' element; defaulting to type='{self._type}", logging.WARN)
        if "description" in all_elements.keys():
            self._description = DataElementSchema._parseDescription(all_elements['description'])
        else:
            self._description = "Unknown"
            Logger.Log(f"{name} EventDataElement config does not have a 'description' element; defaulting to description='{self._description}", logging.WARN)
        if "details" in all_elements.keys():
            self._details = DataElementSchema._parseDetails(details=all_elements['details'])
        else:
            self._details = None
        _leftovers = { key : val for key,val in all_elements.items() if key not in {"type", "description", "details"} }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def AsMarkdown(self) -> str:
        ret_val : str = f"- **{self.Name}** : *{self.ElementType}*, {self.Description}"
        if self.Details is not None:
            detail_markdowns = [f"    - **{name}** - {desc}  " for name,desc in self.Details.items()]
            detail_joined = '\n'.join(detail_markdowns)
            ret_val += f"  \n  Details:  \n{detail_joined}"
        return ret_val

    @property
    def AsMarkdownRow(self) -> str:
        ret_val : str = f"| {self.Name} | {self.ElementType} | {self.Description} |"
        if self.Details is not None:
            detail_markdowns = [f"**{name}** : {desc}" for name,desc in self.Details.items()]
            ret_val += ', '.join(detail_markdowns)
        ret_val += " |"
        return ret_val

    @property
    def ElementType(self) -> str:
        return self._type

    @property
    def Description(self) -> str:
        return self._description

    @property
    def Details(self) -> Optional[Dict[str, str]]:
        return self._details
    
    @staticmethod
    def _parseElementType(event_type):
        ret_val : str
        if isinstance(event_type, str):
            ret_val = event_type
        else:
            ret_val = str(event_type)
            Logger.Log(f"EventDataElement type was not a string, defaulting to str(type) == {ret_val}", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parseDescription(description):
        ret_val : str
        if isinstance(description, str):
            ret_val = description
        else:
            ret_val = str(description)
            Logger.Log(f"EventDataElement description was not a string, defaulting to str(description) == {ret_val}", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parseDetails(details):
        ret_val : Dict[str, str] = {}
        if isinstance(details, dict):
            for key,val in details.items():
                if isinstance(key, str):
                    if isinstance(val, str):
                        ret_val[key] = val
                    else:
                        ret_val[key] = str(val)
                        Logger.Log(f"EventDataElement detail value for key {key} was unexpected type {type(val)}, defaulting to str(val) == {ret_val[key]}", logging.WARN)
                else:
                    _key = str(key)
                    Logger.Log(f"EventDataElement detail key was unexpected type {type(key)}, defaulting to str(key) == {_key}", logging.WARN)
                    if isinstance(val, str):
                        ret_val[_key] = val
                    else:
                        ret_val[_key] = str(val)
                        Logger.Log(f"EventDataElement detail value for key {_key} was unexpected type {type(val)}, defaulting to str(val) == {ret_val[_key]}", logging.WARN)
        else:
            ret_val = {}
            Logger.Log(f"EventDataElement details was not a dict, defaulting to empty dict.", logging.WARN)
        return ret_val



# import libraries
from typing import Any, List
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class EchoSessionID(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._session_id = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["SessionID"]

    def _updateFromEvent(self, event:Event) -> None:
        return

    def _updateFromFeatureData(self, feature:FeatureData):
        self._session_id = feature.FeatureValues[0]

    def _getFeatureValues(self) -> List[Any]:
        return [f"The sess ID is: {self._session_id}"]

    # *** Optionally override public functions. ***



## import standard libraries
from typing import Any, Callable, Dict, List, Optional
## import local files
from . import features
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.games.SHADOWSPECT.features import *
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema

## @class ShadowspectExtractor
#  Extractor subclass for extracting features from Shadowspects game data.
class ShadowspectLoader(GeneratorLoader):

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Feature:
        ret_val : Feature
        match feature_type:
            case "MoveShapeCount":
                ret_val = MoveShapeCount.MoveShapeCount(params=extractor_params)
            case "SessionID":
                ret_val = SessionID.SessionID(params=extractor_params, session_id=self._session_id)
            case "FunnelByUser":
                ret_val = FunnelByUser.FunnelByUser(params=extractor_params)
            case "LevelsOfDifficulty":
                ret_val = LevelsOfDifficulty.LevelsOfDifficulty(params=extractor_params)
            case "SequenceBetweenPuzzles":
                ret_val = SequenceBetweenPuzzles.SequenceBetweenPuzzles(params=extractor_params)
            case "SequenceWithinPuzzles":
                ret_val = SequenceWithinPuzzles.SequenceWithinPuzzles(params=extractor_params)
            case _:
                raise NotImplementedError(f"'{feature_type}' is not a valid feature for Shadowspect.")
        return ret_val

    def _loadDetector(self, detector_type:str, name:str, detector_args:Dict[str,Any], trigger_callback:Callable[[Event], None], count_index:Optional[int] = None) -> Detector:
        raise NotImplementedError(f"'{detector_type}' is not a valid detector for Shadowspect.")

    ## Constructor for the ShadowspectExtractor class.
    def __init__(self, player_id:str, session_id: str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]=None):
        """Constructor for the CrystalLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        self._game_schema = game_schema
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class SequenceLevel(PerLevelFeature):
    def __init__(self, params:GeneratorParameters):
        PerLevelFeature.__init__(self, params=params)
        self._seq = ''
        self._lastSliderType = ''

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.1","CUSTOM.2"]
        # return ["SLIDER_MOVE_RELEASE", "ARROW_MOVE_RELEASE"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        currentSliderType = event.EventData["slider"]
        if currentSliderType != self._lastSliderType:
            sliderTypeCode = ''
            if currentSliderType == "AMPLITUDE": sliderTypeCode = 'a'
            if currentSliderType == "WAVELENGTH": sliderTypeCode = 'w'
            if currentSliderType == "OFFSET": sliderTypeCode = 'o'
            self._seq = self._seq + sliderTypeCode
        self._lastSliderType = currentSliderType

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._seq]

    # *** Optionally override public functions. ***


scenes_list = ["ICE", "VOYAGER", "NOTHING", "EXTREME", "EARTH", "CREDITS"]
task_list = ["VisibleLight","NeutrinoVision","X-ray","xray_vision","pluto1","pluto2","gaze_voyager","gazefeet","gazeup","gazeray","gazeball","gaze_ice","vision_modules","grid","neutrino_event"]
scene_map = {"ICE":
                ["VisibleLight",
                "gazefeet",
                "gazeup"],
            "VOYAGER":
                ["VisibleLight",
                "X-ray",
                "NeutrinoVision"],
            "NOTHING":
                ["VisibleLight"],
            "EXTREME":
                ["VisibleLight",
                "X-ray",
                "NeutrinoVision"],
            "EARTH":
                ["VisibleLight"],
            "CREDITS":
                ["VisibleLight"]
            }
            


""" Lakeland Feature Extractor
Note that a separate file unique to the lakeland extractor is necessary to run this script.
The file is Lakeland Enumerators.json, and is required for the line:
_STR_TO_ENUM = utils.loadJSONFile("games/LAKELAND/Lakeland Enumerators.json")

This json file is created from the fielddaylab/lakeland README on github via
"produce_lakeland_enumerators.py". Please run that script with the appropriate inpath and outpath before running this
script.
"""

## import standard libraries
import logging
import os
import sys
import traceback
import typing
from collections import defaultdict, Counter
from copy import deepcopy
from datetime import datetime, timedelta
from math import sqrt
from pathlib import Path
from typing import Any, Dict, List, Optional
## import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.legacy.LegacyFeature import LegacyFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils import utils
from ogd.core.utils.Logger import Logger

# temp comment

## @class LakelandExtractor
#  Extractor subclass for extracting features from Lakeland game data.
class LakelandExtractor(LegacyFeature):
    _SESS_PREFIX = 'sess_'
    _WINDOW_PREFIX = ''
    _TUTORIAL_MODE = [
        "build_a_house",
        "buy_food",
        "build_a_farm",
        "timewarp",
        "sell_food",
        "buy_fertilizer",
        "buy_livestock",
    ]
    try:
        # TODO: someday, change directory to the directory of current file first, and fix this path, so we don't have indirection issues.
        _STR_TO_ENUM = utils.loadJSONFile(filename="LakelandEnumerators.json", path=Path(".") / "ogd" / "games" / "LAKELAND" / "features" )
    except FileNotFoundError as err:
        Logger.Log(message=f"Could not load Lakeland Enumerators", level=logging.WARNING)
        _STR_TO_ENUM = {}
    _ENUM_TO_STR = {cat: {y: x for x, y in ydict.items()} for cat, ydict in _STR_TO_ENUM.items()}
    _ITEM_MARK_COMBINATIONS = [('food', 'sell'), ('food', 'use'), ('food', 'eat'),
                               ('milk', 'sell'), ('milk', 'use'), ('poop', 'sell'), ('poop', 'use')
                               ]
    _EMOTE_STR_TO_AFFECT = { # 0 neutral, -1 negative, 1 positive
        "null": 0,
        "fullness_motivated_txt": 0,
        "fullness_desperate_txt": -1,
        "energy_desperate_txt": -1,
        "joy_motivated_txt": 0,
        "joy_desperate_txt": -1,
        "puke_txt": -1,
        "yum_txt": 0,
        "tired_txt": -1,
        "happy_txt": 1,
        "swim_txt": 0,
        "sale_txt": 0
    }

    _TUT_NAME_COMPLIANT_BUY_NAME = [
        ("build_a_house", "house"),
        ("buy_food", "food"),
        ("build_a_farm", "farm"),
        ("buy_fertilizer", "fertilizer"),
        ("buy_livestock", "livestock"),
    ]
    _BUILDING_TYPES = ['farm', 'livestock', 'home', 'road', 'sign', 'grave'] # cannot remove these items

    _ACTIVE_EVENTS = ["SELECTTILE", "SELECTFARMBIT", "SELECTITEM", "SELECTBUY", "BUY",
                    "CANCELBUY","TILEUSESELECT","ITEMUSESELECT", 'STARTGAME', 'TOGGLENUTRITION',
                      "TOGGLESHOP", "TOGGLEACHIEVEMENTS", "SKIPTUTORIAL", "SPEED"]

    _EXPLORATORY_EVENTS = ['SELECTTILE', 'SELECTFARMBIT', 'SELECTITEM', 'TOGGLEACHIEVEMENTS', 'TOGGLESHOP', 'TOGGLENUTRITION']

    _IMPACT_EVENTS = ["BUY","TILEUSESELECT","ITEMUSESELECT"]

    _NULL_FEATURE_VALS = ['null', 0, None]


    ## Constructor for the LakelandExtractor class.
    #  Initializes some custom private data (not present in base class) for use
    #  when calculating some features.
    #  Sets the sessionID feature.
    #
    #  @param session_id The id number for the session whose data is being processed
    #                    by this extractor instance.
    #  @param game_table A data structure containing information on how the db
    #                    table associated with this game is structured.
    #  @param game_schema A dictionary that defines how the game data itself is
    #                     structured.
    def __init__(self, params:GeneratorParameters, game_schema:GameSchema, session_id:str):
        # Initialize superclass
        super().__init__(params=params, game_schema=game_schema, session_id=session_id)
        # Set window and overlap size
        self._NUM_SECONDS_PER_WINDOW = game_schema.Config[LakelandExtractor._WINDOW_PREFIX+'WINDOW_SIZE_SECONDS']
        self._NUM_SECONDS_PER_WINDOW_OVERLAP = game_schema.Config["WINDOW_OVERLAP_SECONDS"]
        self._GAME_SCHEMA = game_schema
        self._IDLE_THRESH_SECONDS = game_schema.Config['IDLE_THRESH_SECONDS']
        self.WINDOW_RANGE = range(game_schema.LevelRange.stop)
        self._WINDOW_RANGES = self._get_window_ranges()
        self._cur_gameplay = 1
        self._startgame_count = 0
        self.debug_strs = []

        # set window range
        
        self.reset()
        self.setValByName('num_play', self._cur_gameplay)
    
    def _extractFeaturesFromEvent(self, event:Event):
        try:
            self._updateFromEvent(event)
        except Exception as e:
            if len(self.debug_strs) > 10:
                debug_strs = self.debug_strs[:5] + ['...'] + self.debug_strs[-5:]
            else:
                debug_strs = self.debug_strs
            exc_type, exc_value, exc_traceback = sys.exc_info()
            Logger.Log('\n'.join(debug_strs), logging.DEBUG)
            Logger.Log('\n'.join(traceback.format_exception(exc_type, exc_value, exc_traceback)), logging.ERROR)
            if True:
                pass


    ## Function to perform extraction of features from a row.
    #
    #  @param row_with_complex_parsed A row of game data from the db, with the
    #                                 "complex data" already parsed from JSON.
    #  @param table_schema A data structure containing information on how the db
    #                      table assiciated with this game is structured.
    def _updateFromEvent(self, event:Event):

        # put some data in local vars, for readability later.
        self.event_client_time = event.Timestamp
        server_time = event.EventData['server_time']
        event_type_str = LakelandExtractor._ENUM_TO_STR['EVENT CATEGORIES'][int(event.EventName.split('.')[-1])].upper()
        if self._end and event_type_str in ['STARTGAME', 'NEWFARMBIT']:
            self._cur_gameplay += 1
            self.setValByName('num_play', self._cur_gameplay)
        if not self._VERSION:
            self._VERSION = event.LogVersion
            self.setValByName("version", self._VERSION)
        # Check for invalid row.
        if self.ExtractionMode == ExtractionMode.SESSION and event.SessionID != self._session_id:
            Logger.Log(f"Got a row with incorrect session id! Expected {self._session_id}, got {event.SessionID}!", logging.ERROR)
        # If row is valid, process it.
        else:
            # If we haven't set persistent id, set now.
            if self.getValByName(feature_name="persistentSessionID") == 0:
                self.setValByName(feature_name="persistentSessionID", new_value=event.UserData['persistent_session_id'])
            # if self.getValByName(feature_name="player_id") == 0:
            #     self.setValByName(feature_name="player_id",
            #                                new_value=row_with_complex_parsed[table_schema.player_id_index])

            # if this is the first row
            if not self._CLIENT_START_TIME:
                # initialize this time as the start
                self._extractFromFirst(event.Timestamp, event.EventData)

            # set current windows
            new_windows = self._get_windows_at_time(event.Timestamp)
            if not new_windows:
                return # reached past the max windows
            old_max, new_min = max(self._cur_windows), min(new_windows)
            skipped_windows = list(range(old_max+1, new_min))
            started_windows = [w for w in new_windows if w not in self._cur_windows]
            finished_windows = [w for w in self._cur_windows if w not in new_windows]
            window_len_td = timedelta(self._NUM_SECONDS_PER_WINDOW)
            if skipped_windows:
                if self._active:
                    self._add_event("Z")
                    self.feature_count('count_idle', windows=[skipped_windows[0]])
                self._active = False
                self.setValByName('rt_currently_active', int(self._active))
                for w in skipped_windows:
                    self.start_window(w)
                    self.feature_inc('time_in_secs', window_len_td, windows=[w])
                    self.feature_inc('time_idle', window_len_td, windows=[w])
                    self.finish_window(w)

            for w in started_windows:
                self.start_window(w)
                # prev_window_end_time = self._CLIENT_START_TIME + \
                #     timedelta(seconds = self._WINDOW_RANGES[w-1][1],
                #                         milliseconds= 999)
                # if event_type_str not in LakelandExtractor._ACTIVE_EVENTS:
                #     self._extractFromActive(prev_window_end_time, None)
                # else:
                #     self._extractFromInactive(prev_window_end_time, None)
            for w in finished_windows:
                self.finish_window(w)

            self._cur_windows = new_windows

            # Record that an event of any kind occurred, for the window & session
            time_since_start = self.time_since_start(event.Timestamp)
            self.feature_count(feature_base="EventCount")
            self.setValByName(feature_name="sessDuration", new_value=time_since_start)
            if event_type_str == "BUY" and event.EventData["success"]:
                debug_str = LakelandExtractor._ENUM_TO_STR['BUYS'][event.EventData["buy"]].upper()
            if event_type_str == 'STARTGAME':
                debug_str = "START" if event.EventData.get("continue") == False else "CONTINUE"
                debug_str += f' Language: {event.EventData.get("language")}'
            else:
                debug_str = ''
            self.add_debug_str(f'{self._num_farmbits } {time_since_start} {event_type_str} {debug_str}')


            self._extractFromAll(event.Timestamp, event.EventData)
            # Then, handle cases for each type of event
            if event_type_str in LakelandExtractor._ACTIVE_EVENTS:
                self._extractFromActive(event.Timestamp, event.EventData)
            else:
                self._extractFromInactive(event.Timestamp, event.EventData)
            if event_type_str in LakelandExtractor._EXPLORATORY_EVENTS:
                self._extractFromExplore(event.Timestamp, event.EventData)
            if event_type_str in LakelandExtractor._IMPACT_EVENTS:
                self._extractFromImpact(event.Timestamp, event.EventData)
            match event_type_str:
                case "GAMESTATE":
                    self._extractFromGamestate(event.Timestamp, event.EventData)
                case "STARTGAME":
                    self._extractFromStartgame(event.Timestamp, event.EventData)
                case "CHECKPOINT":
                    self._extractFromCheckpoint(event.Timestamp, event.EventData)
                case "SELECTTILE":
                    self._extractFromSelecttile(event.Timestamp, event.EventData)
                case "SELECTFARMBIT":
                    self._extractFromSelectfarmbit(event.Timestamp, event.EventData)
                case "SELECTITEM":
                    self._extractFromSelectitem(event.Timestamp, event.EventData)
                case "SELECTBUY":
                    self._extractFromSelectbuy(event.Timestamp, event.EventData)
                case "BUY":
                    self._extractFromBuy(event.Timestamp, event.EventData)
                case "CANCELBUY":
                    self._extractFromCancelbuy(event.Timestamp, event.EventData)
                case "ROADBUILDS":
                    self._extractFromRoadbuilds(event.Timestamp, event.EventData)
                case "TILEUSESELECT":
                    self._extractFromTileuseselect(event.Timestamp, event.EventData)
                case "ITEMUSESELECT":
                    self._extractFromItemuseselect(event.Timestamp, event.EventData)
                case "TOGGLENUTRITION":
                    self._extractFromTogglenutrition(event.Timestamp, event.EventData)
                case "TOGGLESHOP":
                    self._extractFromToggleshop(event.Timestamp, event.EventData)
                case "TOGGLEACHIEVEMENTS":
                    self._extractFromToggleachievements(event.Timestamp, event.EventData)
                case "SKIPTUTORIAL":
                    self._extractFromSkiptutorial(event.Timestamp, event.EventData)
                case "SPEED":
                    self._extractFromSpeed(event.Timestamp, event.EventData)
                case "ACHIEVEMENT":
                    self._extractFromAchievement(event.Timestamp, event.EventData)
                case "FARMBITDEATH":
                    self._extractFromFarmbitdeath(event.Timestamp, event.EventData)
                case "BLURB":
                    self._extractFromBlurb(event.Timestamp, event.EventData)
                case "CLICK":
                    self._extractFromClick(event.Timestamp, event.EventData)
                case "RAINSTOPPED":
                    self._extractFromRainstopped(event.Timestamp, event.EventData)
                case "HISTORY":
                    self._extractFromHistory(event.Timestamp, event.EventData)
                case "ENDGAME":
                    self._extractFromEndgame(event.Timestamp, event.EventData)
                case "EMOTE":
                    self._extractFromEmote(event.Timestamp, event.EventData)
                case "FARMFAIL":
                    self._extractFromFarmfail(event.Timestamp, event.EventData)
                case "BLOOM":
                    self._extractFromBloom(event.Timestamp, event.EventData)
                case "FARMHARVESTED":
                    self._extractFromFarmharvested(event.Timestamp, event.EventData)
                case "MILKPRODUCED":
                    self._extractFromMilkproduced(event.Timestamp, event.EventData)
                case "POOPPRODUCED":
                    self._extractFromPoopproduced(event.Timestamp, event.EventData)
                case "DEBUG":
                    self._extractFromDebug(event.Timestamp, event.EventData)
                case "NEWFARMBIT":
                    self._extractFromNewfarmbit(event.Timestamp, event.EventData)
                case "AVAILABLEFOOD":
                    self._extractFromAvailablefood(event.Timestamp, event.EventData)
                case "MONEYRATE":
                    self._extractFromMoneyrate(event.Timestamp, event.EventData)
                case "SADFARMBITS":
                    self._extractFromSadfarmbits(event.Timestamp, event.EventData)
                case "LAKENUTRITION":
                    self._extractFromLakenutrition(event.Timestamp, event.EventData)
                case "SALESTART":
                    self._extractFromSalestart(event.Timestamp, event.EventData)
                case "SALEEND":
                    self._extractFromSaleend(event.Timestamp, event.EventData)
                case "RAINSTARTED":
                    self._extractFromRainstarted(event.Timestamp, event.EventData)
                case "EATFOOD":
                    self._extractFromEatfood(event.Timestamp, event.EventData)
                case "RESET":
                    self._extractFromReset(event.Timestamp, event.EventData)
                case _:
                    raise Exception(f"Found an unrecognized event type: {event.EventName}")

    def _extractFromFirst(self, event_client_time, event_data):
        self._CLIENT_START_TIME = event_client_time
        self._last_speed_change = self._CLIENT_START_TIME
        self._last_active_event_time = event_client_time
        self._last_event_time = event_client_time
        self._active = True
        self.add_debug_str(f'{"*" * 10}\nPlay {self._cur_gameplay} \n' +
                           f'{self._session_id} v{self._VERSION} @ {self._CLIENT_START_TIME}')
        self.setValByName("play_year", self._CLIENT_START_TIME.year)
        self.setValByName("play_month", self._CLIENT_START_TIME.month)
        self.setValByName("play_day", self._CLIENT_START_TIME.day)
        self.setValByName("play_hour", self._CLIENT_START_TIME.hour)
        self.setValByName("play_minute", self._CLIENT_START_TIME.minute)
        self.setValByName("play_second", self._CLIENT_START_TIME.second)


    def _extractFromAll(self, event_client_time, event_data):
        pass

    def _extractFromActive(self, event_client_time, event_data=None):
        # if event_data_complex is none, this isnt an event, its a start/end of a window
        zero_td = timedelta(0)
        idle_thresh_td = timedelta(seconds=self._IDLE_THRESH_SECONDS)
        time_between_actives = event_client_time - self._last_active_event_time
        if time_between_actives < idle_thresh_td:
            self._increment_sess_feature('sess_time_active', time_between_actives)
        else:
            self._increment_sess_feature('sess_time_idle', time_between_actives)
            pass
        if len(self._cur_windows) == 1:
            prev_windows = self._get_windows_at_time(self._last_active_event_time)
            if len(prev_windows) == 1:
                # only works if windows are only in series, never in parallel
                # feature to increment
                prev_window, cur_window = prev_windows[0], self._cur_windows[0]
                cur_window_start_time, _ = self._get_window_start_end_time(cur_window)
                if self._last_active_event_time < cur_window_start_time:
                    # split between two windows
                    time_between_actives_1 = cur_window_start_time - self._last_active_event_time
                    time_between_actives_2 = event_client_time - cur_window_start_time
                    if time_between_actives < idle_thresh_td:
                        # active, two windows
                        self._increment_feature_in_cur_windows('time_active', time_between_actives_1, windows=[prev_window])
                        self._increment_feature_in_cur_windows('time_active', time_between_actives_2, windows=[cur_window])
                    else:
                        # idle, two windows
                        if not self._active:
                            # if went idle
                            self._add_event("Z")
                            self.feature_count('count_idle', windows=[prev_window])
                            self._active = False
                        self._increment_feature_in_cur_windows('time_idle', time_between_actives_1, windows=[prev_window])
                        self._increment_feature_in_cur_windows('time_idle', time_between_actives_2, windows=[cur_window])
                else:
                    # keep in one window
                    if time_between_actives < idle_thresh_td:
                        # active, one window
                        self._increment_feature_in_cur_windows('time_active', time_between_actives)
                    else:
                        # idle, one window
                        if not self._active:
                            # if went idle
                            self._add_event("Z")
                            self.feature_count('count_idle')
                            self._active = False
                        self._increment_feature_in_cur_windows('time_idle', time_between_actives)



        #     # if all in one window
        #     prev_windows = self._get_windows_at_time(self._last_active_event_time)
        #     # logic breaks if multiple windows at a time
        #     assert len(prev_windows) == 1
        #
        #     # if split between multiple prev window
        #     if time_between_actives > timedelta(seconds=self._IDLE_THRESH_SECONDS):
        #         self._add_event("Z")
        #         self.feature_count('count_idle')
        #         time_idle = time_between_actives
        #         time_active = zero_td
        #     else:
        #         time_active = time_between_actives
        #         time_idle = zero_td
        # else:
        #     time_active = zero_td
        #     time_idle = event_client_time - self._last_event_time

        self._last_active_event_time = event_client_time
        self._last_event_time = event_client_time
        self._active = True
        # self.feature_inc('time_in_secs',time_active+time_idle)
        # self.feature_inc('time_idle', time_idle)
        # self.feature_inc('time_active', time_active)
        self.setValByName('rt_currently_active',int(self._active))
        if event_data is not None:
            self.feature_count(feature_base="ActiveEventCount")


    def _extractFromInactive(self, event_client_time, event_data=None):
        # if event_data_complex is none, this isnt an event, its a start/end of a window
        # zero_td = timedelta(0)
        # if not self._active:
        #     time_idle = event_client_time - self._last_event_time
        # else:
        #     time_since_last_active = event_client_time - self._last_active_event_time
        #     idle_thresh = timedelta(seconds=self._IDLE_THRESH_SECONDS)
        #     if time_since_last_active > idle_thresh:
        #         self._active = False
        #         self._add_event("Z")
        #         self.feature_count('count_idle')
        #         time_idle = time_since_last_active
        #     else:
        #         time_idle = zero_td
        #
        # self._last_event_time = event_client_time
        # self.feature_inc('time_in_secs',time_idle)
        # self.feature_inc('time_idle', time_idle)
        self.setValByName('rt_currently_active', int(self._active))
        if event_data is not None:
            self.feature_count(feature_base="InactiveEventCount")

    def _extractFromExplore(self, event_client_time, event_data):
        self.feature_count('count_explore_events')

    def _extractFromImpact(self, event_client_time, event_data):
        self.feature_count('count_impact_events')



    def _extractFromGamestate(self, event_client_time, event_data):
        pass
        # # assign event_data variables
        # d = event_data
        # _tiles = d["tiles"]
        # _farmbits = d["farmbits"]
        # _items = d["items"]
        # _money = d["money"]
        # _speed = d["speed"]
        # _achievements = d["achievements"]
        # _num_checkpoints_completed = d["num_checkpoints_completed"]
        # _raining = d["raining"]
        # _curr_selection_type = d["curr_selection_type"]
        # _curr_selection_data = d.get("curr_selection_data")
        # _camera_center = d["camera_center"]
        # _gametime = d["gametime"]
        # _timestamp = d["timestamp"]
        # _num_food_produced = d["num_food_produced"]
        # _num_milk_produced = d["num_milk_produced"]
        # _num_poop_produced = d["num_poop_produced"]
        #
        # # reformat raw variable functions
        # def array_to_mat(num_columns, arr):
        #     assert len(arr) % num_columns == 0
        #     return [arr[i:i + num_columns] for i in range(0, len(arr), num_columns)]
        # def read_stringified_array(arr):
        #     if not arr:
        #         return []
        #     return [int(x) for x in arr.split(',')]
        #
        # # reformat array variables
        # _tiles = read_stringified_array(_tiles)
        # _farmbits = read_stringified_array(_farmbits)
        # _items = read_stringified_array(_items)
        # _tiles = array_to_mat(4, _tiles)
        # _farmbits = array_to_mat(9, _farmbits)
        # _items = array_to_mat(4, _items)
        #
        # # set class variables
        # if self._first_gamestate and self._continue:
        #     for i,t in enumerate(_tiles):
        #         type = t[3]
        #         if LakelandExtractor._ENUM_TO_STR['TILE TYPE'][type] in LakelandExtractor._BUILDING_TYPES:
        #             self.building_xys.append(tile_i_to_xy(i))
        #         if type in [9,10]: #farm or livestock
        #             self._tiles[tile_i_to_xy(i)]['type'] = type
        #
        # if self._num_farmbits != len(_farmbits):
        #     # if the num farmbits arent as expected or just one more than expected (sometimes takes a while to show), show warning.
        #     self.log_warning(f'Gamestate showed {len(_farmbits)} but expected {self._num_farmbits} farmbits.')
        #      # self._change_population(event_client_time, len(_farmbits) - self._num_farmbits)
        #
        #
        # #  feature functions
        # # def farms_low_productivity(tiles):
        # #     return [t for t in tiles if t[3] == 9 and t[1] < 3]  # type == farm and nutrition < 3 (the circle is heavily black)
        #
        # 0# def lake_tiles_bloom(tiles):
        # #     return [t for t in tiles if t[3] == 5 and t[1] > 25]  # lake tiles bloom when nutrition > 10% of 255,
        # # so technically not all tiles > 25 nutrition are blooming, but we will put the cutoff here
        #
        # def avg_lake_nutrition(tiles):
        #     lake_nutritions = [t[1] for t in tiles if t[2] == 5] # 5 is the lake enum for tile type
        #     if len(lake_nutritions) == 0:
        #         self.log_warning(f'No lake nutritions!!')
        #
        #     return sum(lake_nutritions) / len(lake_nutritions)
        #
        # def num_items_in_play(items, tiles):
        #     cur_num_items_on_screen = LakelandExtractor.onscreen_item_dict()
        #     for it in items:
        #         type = it[2]
        #         key = LakelandExtractor._ENUM_TO_STR["ITEM TYPE"][type]
        #         cur_num_items_on_screen[key] += 1
        #     for t in tiles:
        #         type = t[3]
        #         key = LakelandExtractor._ENUM_TO_STR["TILE TYPE"][type]
        #         if key in cur_num_items_on_screen:
        #             cur_num_items_on_screen[key] += 1
        #     return cur_num_items_on_screen
        #
        # # set features
        # self.feature_count('count_gamestate_logs')
        #
        # for (key, cur_num) in num_items_in_play(_items, _tiles).items():
        #     self.feature_max_min(f'num_{key}_in_play', cur_num)
        # self.feature_max_min(fname_base="avg_lake_nutrition", val=avg_lake_nutrition(_tiles))
        # # self.feature_max_min("num_farms_low_productivity", len(farms_low_productivity(_tiles)))
        # # self.feature_max_min("num_lake_tiles_in_bloom", len(lake_tiles_bloom(_tiles)))

    def _extractFromStartgame(self, event_client_time, event_data):
        self._startgame_count += 1
        if self._startgame_count > 1:
            self.add_debug_str(f"Starting game {self._startgame_count}.")
        if self._startgame_count != self._cur_gameplay:
            self.log_warning(f"Starting game {self._cur_gameplay} but this is the {self._startgame_count}th start!")

        # assign event_data variables
        d = event_data
        _tile_states = d["tile_states"]
        _tile_nutritions = d["tile_nutritions"]
        _continue = d["continue"]
        _language = d["language"]
        _audio = d["audio"]
        _fullscreen = d["fullscreen"]

        # set class variables
        self._TILE_OG_TYPES = _tile_states
        self._continue = _continue
        self._first_gamestate = False


        # feature functions

        # set features
        self.setValByName("continue", _continue)
        self.setValByName("language", _language)
        self.setValByName("audio", _audio)
        self.setValByName("fullscreen", _fullscreen)

        if not self._continue:
            for key in self.onscreen_item_dict():
                self.feature_max_min(f'num_{key}_in_play', 0)



    def _extractFromCheckpoint(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _event_category = d["event_category"] # start or end
        _event_label = d["event_label"] # tutorial name
        _event_type = d["event_type"]
        _blurb_history = d["blurb_history"]
        _client_time = d["client_time"]


        # reformat variables
        _blurb_history = self.reformat_history_array(_blurb_history, _client_time)

        # helpers
        assert _event_type == 'tutorial'
        is_tutorial_end = 1 if _event_category == 'end' else 0
        is_tutorial_start = not is_tutorial_end
        time_since_start = self.time_since_start(event_client_time)
        if is_tutorial_end:
            assert _blurb_history is not None

        # set class variables
        # self._tutorial_start_and_end_times[_event_label][is_tutorial_end] = event_client_time
        if is_tutorial_end:
            self._tutorial_times_per_blurb[_event_label] = list_deltas(_blurb_history)

        # feature functions
        def avg_time_per_blurb(tut_name):
            time_per_blurb = self._tutorial_times_per_blurb.get(tut_name)
            assert time_per_blurb
            return avg(time_per_blurb)

        # set features
        if is_tutorial_start:
            self.feature_time_since_start(feature_base=f"time_to_{_event_label}_tutorial")
            self.feature_inc(feature_base='count_encounter_tutorial', increment=is_tutorial_start)
        # TODO: Case where blurb history exists but player skipped
        if is_tutorial_end and len(_blurb_history) > 1:
            self.setValByName(
                feature_name=f"{LakelandExtractor._SESS_PREFIX}avg_time_per_blurb_in_{_event_label}_tutorial",
                new_value=avg_time_per_blurb(_event_label))

    def _extractFromSelecttile(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _tile = d["tile"]
        _marks = d["marks"]

        # set features
        self.feature_count("count_inspect_tile")

    def _extractFromSelectfarmbit(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _farmbit = d["farmbit"]

        # set features
        self.feature_count("count_inspect_farmbit")

    def _extractFromSelectitem(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _item = d["item"]

        # set features
        self.feature_count("count_inspect_item")

    def _extractFromSelectbuy(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _buy = d["buy"]
        _cost = d["cost"]
        _cur_money = d["curr_money"]
        _success = d["success"]

        # set class variables
        self._selected_buy_cost = _cost
        self._cur_money = _cur_money

    def _extractFromBuy(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _buy = d["buy"]
        _tile = d["tile"]
        _success = d["success"]
        _buy_hovers = d["buy_hovers"]
        _client_time = d["client_time"]

        if not _success:
            return

        # helpers
        valid_hovers = [h for h in _buy_hovers if h[6] > 0] # h[6] (placement_valid) denotes if players can build there or not
        buy_name = LakelandExtractor._ENUM_TO_STR['BUYS'][_buy]

        # set class variables
        if buy_name in LakelandExtractor._BUILDING_TYPES:
            self.building_xys.append(get_tile_txy(_tile))

        if buy_name in ['farm', 'livestock', 'home', 'road', 'sign']:
            self._tiles[get_tile_txy(_tile)]['type'] = buy_name

        # feature functions
        def chose_highest_nutrient_tile(buy_hovers, chosen_tile):
            if not buy_hovers:
                return None
            building_buildable_func = self.get_building_buildable_function() if Event.CompareVersions(self._VERSION, "15") < 0 \
                else lambda buy_t: buy_t[-2] # whether the tile is buildable (v15 buy hover data contains:
                                             # [tile_data_short, buildable, hover_time]
            for t in buy_hovers:
                # t has the current type of land (only type that farms can be built on) and a building (farm) can be placed there
                if t[3] == 1 and building_buildable_func(t):
                    # if any tile found has a nutrient value more than 2 above the chosen tile, fail
                    if t[1] > (chosen_tile[1]+2):  # comparing the nutritions give or take 2/255 ~ 1% of nutrition
                       return 0  # failed to put on highest nutrition tile
            return 1

        # I think that the calculation was mostly fine, but it definitely doesnt work if player never hovers tiles.
        def chose_lowest_nutrient_farm(buy_hovers, chosen_tile):
            if not buy_hovers:
                return None
            for t in buy_hovers:
                if t[3] == 9:  # t has the current type of farm (only type that fertilizer can be built on)
                    # if any tile found has a nutrient value less than 2 below the chosen tile, fail
                    if t[1] < (chosen_tile[1]-2):  # comparing the nutritions give or take 2/255 ~ 1% of nutrition
                        return 0  # failed to put on lowest nutrition tile
            return 1

        def dist_to_lake(tile):
            tile_pt = (tile[4], tile[5])
            lake_xys = self.get_tile_types_xys(['lake'])
            if not lake_xys:
                self.log_warning('There were no lake tiles found!')
                return None
            min_lake_dist = sqrt(min([distance(tile_pt, lake_pt) for lake_pt in lake_xys]))
            return min_lake_dist

        def avg_distance_between_buildings():
            if len(self.building_xys) < 2:
                return 0
            distances = [distance(pt1, pt2) for pt1 in self.building_xys for pt2 in self.building_xys if pt1 != pt2]
            return sum(distances)/len(distances)

        # set features
        self.feature_inc(feature_base="money_spent", increment=self._selected_buy_cost)
        self.feature_time_since_start(f'time_to_first_{buy_name}_buy')
        self.feature_inc(feature_base=f"money_spent_{buy_name}", increment=self._selected_buy_cost)
        self.feature_count(feature_base=f'count_buy_{buy_name}')
        self.feature_average(fname_base=f'avg_num_tiles_hovered_before_placing_{buy_name}', value=len(valid_hovers))
        if buy_name == 'farm':
            self.feature_average(fname_base='percent_building_a_farm_on_highest_nutrition_tile',
                             value=chose_highest_nutrient_tile(valid_hovers, _tile))
        if buy_name == 'fertilizer':
            self.feature_average(fname_base="percent_placing_fertilizer_on_lowest_nutrient_farm",
                                 value=chose_lowest_nutrient_farm(valid_hovers, _tile))
            dist = dist_to_lake(_tile)
            if dist:
                self.feature_average("avg_distance_between_poop_placement_and_lake", dist)
        if buy_name in LakelandExtractor._BUILDING_TYPES:
            self.feature_average(fname_base='avg_avg_distance_between_buildings', value=avg_distance_between_buildings())
            self.feature_max_min(f'num_{buy_name}_in_play', len([t for t in self._tiles.values() if t['type'] == buy_name]))

        self._add_event(_buy)

    # def _tutorial_compliance_check(self, tut_name, event_client_time):
    #     tutorial_end_time = self._tutorial_start_and_end_times[tut_name][1]
    #     if not self._tutorial_compliance[tut_name] and tutorial_end_time:
    #         compliance_interval = event_client_time - tutorial_end_time
    #         fname = f"{tut_name}_compliance_interval"
    #         self.setValByName(feature_name=fname, new_value=compliance_interval)

    def _extractFromCancelbuy(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _selected_buy = d["selected_buy"]
        _cost = d["cost"]
        _buy_hovers = d["buy_hovers"]
        _client_time = d["client_time"]

        # set class variables
        self._selected_buy_cost = 0

    def _extractFromRoadbuilds(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _road_builds = d["road_builds"]
        _client_time = d["client_time"]

    def _extractFromTileuseselect(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _tile = d["tile"]
        _marks = d["marks"]

        # helpers
        mark_change = self._tiles[(get_tile_txy(_tile))]['marks'] != _marks
        if not mark_change:
            return
        self._tiles[(get_tile_txy(_tile))]['marks'] = _marks
        all_farm_marks = [t["marks"] for t in self._tiles.values() if t["type"] == 'farm'] # 9 == farm
        all_dairy_marks = [t["marks"] for t in self._tiles.values() if t["type"] == 'livestock'] # 10 == livestock
        food_uses = [t[0] for t in all_farm_marks] + [t[1] for t in all_farm_marks] # mark index=0,1 food mark
        milk_uses = [t[0] for t in all_dairy_marks] # mark index=0 is the dairy mark
        poop_uses = [t[1] for t in all_dairy_marks] # mark index=1 is the poop mark

        # set class variables
        # self._tiles[get_tile_txy(_tile)]['type'] = _tile[3] # tile index 3 is type

        # feature setters

        self.feature_count(feature_base="count_change_tile_mark")
        self.feature_time_since_start("time_to_first_tilemarkchange")
        for produced, produced_uses in [('food', food_uses), ('milk', milk_uses), ('poop', poop_uses)]:
            mark_counts = {LakelandExtractor._ENUM_TO_STR["MARK"][k]: v for k, v in Counter(produced_uses).items()}
            # for each produced item (food, milk, poop), we now have mark counts
            # {eat: ?, use: ?, sell: ?} of total numbers marked eat, use, sell.
            for use, count in mark_counts.items():
                self.feature_max_min(f"num_{produced}_marked_{use}", count) #mark use = eat in case of food
                self.feature_max_min(f"num_per_capita_{produced}_marked_{use}", count / (self._num_farmbits or 1))
                self.feature_max_min(f"percent_{produced}_marked_{use}", count/len(produced_uses))

    def _extractFromItemuseselect(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _item = d["item"]
        _prev_mark = d["prev_mark"]

        # feature setters
        if d["prev_mark"] != _item[3]: # if the mark has changed
            self.feature_count(feature_base="count_change_item_mark")
            self.feature_time_since_start("time_to_first_itemmarkchange")

    def _extractFromTogglenutrition(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _to_state = d["to_state"]
        _tile_nutritions = d["tile_nutritions"]

        # set class variables
        self._nutrition_view_on = _to_state # (on is state==1)
        if self._nutrition_view_on:
            self._last_nutrition_open = event_client_time
            self.feature_time_since_start("time_to_first_togglenutrition")

        # set features
        self.set_time_in_nutrition_view(event_client_time)

    def _extractFromToggleshop(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _shop_open = d["shop_open"]

        # set features
        if _shop_open:
            self.feature_count("count_open_shop")

    def _extractFromToggleachievements(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _achievements_open = d["achievements_open"]

        # set features
        if _achievements_open:
            self.feature_count("count_open_achievements")
            self.feature_time_since_start("time_to_first_toggleachievements")

    def _extractFromSkiptutorial(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data

        # set features
        self.feature_count("count_skips")
        self.feature_time_since_start("time_to_first_skip")

    def _extractFromSpeed(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _cur_speed = d["cur_speed"]
        _prev_speed = d["prev_speed"]
        _manual = d["manual"]

        # set class variables
        self._cur_speed = _cur_speed


        # set features (this feature is modified by 2+ events)
        if self._cur_speed != _prev_speed:
            self.update_time_at_speed(event_client_time, _prev_speed)
            if LakelandExtractor._ENUM_TO_STR["SPEED"][_cur_speed] == "pause":
                self._add_event("P")
            elif LakelandExtractor._ENUM_TO_STR["SPEED"][_cur_speed] == "play":
                self._add_event("Q")
            elif LakelandExtractor._ENUM_TO_STR["SPEED"][_cur_speed] == "fast":
                self._add_event("R")
            elif LakelandExtractor._ENUM_TO_STR["SPEED"][_cur_speed] == "vfast":
                self._add_event("S")

    def _extractFromAchievement(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _achievement = d["achievement"]

        # helpers
        achievement_str = LakelandExtractor._ENUM_TO_STR["ACHIEVEMENTS"][_achievement]

        # set features
        self.feature_count("count_achievements")
        self.feature_time_since_start(f"time_to_{achievement_str}_achievement")

    def _extractFromFarmbitdeath(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _farmbit = d["farmbit"]
        _grave = d["grave"]

        # set session variables
        self._change_population(event_client_time, -1)

        # set features
        self.feature_count("count_deaths")
        self.feature_time_since_start('time_to_first_death')
        self._add_event("D")

    def _extractFromBlurb(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data

    def _extractFromClick(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data

    def _extractFromRainstopped(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        self.feature_count("count_rains")
        self.feature_time_since_start("time_to_first_rain")

    def _extractFromHistory(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _client_time = d["client_time"]
        _camera_history = d["camera_history"]
        _emote_history = d["emote_history"]

        # reformat variables
        # _emote_history = self.reformat_history_array(_emote_history, _client_time)

        # set global variables
        # for e in _emote_history:
        #     self._add_emote_to_window_by_time(e)

    def _extractFromEndgame(self, event_client_time, event_data):
        # assign event_data variables 
        d = event_data
        # (none) = d["(none)"]
        self.update_time_at_speed(event_client_time, self._cur_speed)
        if self._nutrition_view_on:
            self.set_time_in_nutrition_view(event_client_time)

        self._end = True

    def _extractFromEmote(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _farmbit = d["farmbit"]
        _emote_enum = d["emote_enum"]

        # fix bug in v13 logging where emote 2s show up as emote 0s
        if Event.CompareVersions(self._VERSION, "13") == 0 and _emote_enum == 0:
            _emote_enum = 2

        # helpers
        emote_str = LakelandExtractor._ENUM_TO_STR['EMOTES'][_emote_enum]
        affect = LakelandExtractor._EMOTE_STR_TO_AFFECT[emote_str]
        self.feature_time_since_start(f"time_to_first_{emote_str.replace('_txt','')}_emote")
        is_positive = affect == 1
        is_neutral = affect == 0
        is_negative = affect == -1
        if(self._num_farmbits) < 1:
            # self.WarningMessage(f"Num farmbits < 1 @ {self._num_farmbits}!! Setting to 1.")
            self.log_warning(f"Num farmbits < 1 @ {self._num_farmbits}!! Setting to 1.")
            self._num_farmbits = 1

        per_capita_val = 1/self._num_farmbits

        # set features
        self.feature_inc(f"count_{emote_str}_emotes_per_capita", per_capita_val)

        self.feature_average("percent_negative_emotes", is_negative)
        self.feature_average("percent_neutral_emotes", is_neutral)
        self.feature_average("percent_positive_emotes", is_positive)

        self.feature_inc("total_negative_emotes", is_negative)
        self.feature_inc("total_neutral_emotes", is_neutral)
        self.feature_inc("total_positive_emotes", is_positive)

    def _extractFromFarmfail(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _tile = d["tile"]
        _marks = d["marks"]

        # helpers

        # set class variables

        # set features
        self.feature_count("count_farmfails")
        self.feature_time_since_start("time_to_first_farmfail")

    def _extractFromBloom(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _tile = d["tile"]
        _marks = d["marks"]

        # helpers
        # set class variables
        # set features
        self.feature_count("count_blooms")
        self.feature_time_since_start('time_to_first_bloom')
        self._add_event("B")

    def _extractFromFarmharvested(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _tile = d["tile"]
        _marks = d["marks"]

        # helpers
        # set class variables
        # set features
        self.feature_inc("count_food_produced", 2)

    def _extractFromMilkproduced(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _tile = d["tile"]
        _marks = d["marks"]

        # helpers
        # set class variables
        # set features
        self.feature_inc("count_milk_produced", 1)
        self.feature_time_since_start("time_to_first_milkproduced")

    def _extractFromPoopproduced(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _tile = d["tile"]
        _marks = d["marks"]

        # helpers
        # set class variables
        # set features
        self.feature_inc("count_poop_produced", 1)
        self.feature_time_since_start("time_to_first_poopproduced")

    def _extractFromDebug(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        self.add_debug_str('Set debug')

        # helpers
        self._debug = True
        # set class variables
        # set features
        self.setValByName(feature_name='debug', new_value=1)

    def _extractFromNewfarmbit(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _farmbit = d["farmbit"]

        # helpers
        # set class variables
        self._change_population(client_time=event_client_time, increment=1)
        # set features

    def _extractFromAvailablefood(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _foods = d["food"]  # gg.food | Current amount of food available on board  |
        _farmbits = d["farmbit"]  # gg.farmbits.length | Current number of farmbits on board  |
        _foodrate = d["food_perfarmbit"]  # gg.food/gg.farmbits.length | Food available per farmbit  |

        #helpers
        # set class variables
        # set features

    def _extractFromMoneyrate(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _money = d["money"]  # sell_rate*item_worth_food | Money accumulated according to amount of available items to sell  |
        _rate = d["rate"]  # money/permin | Rate of money accumulation  |

        #helpers
        # set class variables
        # set features

    def _extractFromSadfarmbits(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _sads = d["sad"]  # sad | Number of farmbits with joy status of DESPERATE  |
        _farmbits = d["farmbit"]  # gg.farmbits.length | Current number of farmbits on board  |
        _sad_perfarmbit = d["sad_perfarmbit"]  # sad/gg.farmbits.length | Amount of sadness per farmbit  |

        #helpers
        # set class variables
        # set features

    def _extractFromLakenutrition(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _lake_pos_tile = d["lake_pos_tile"]  # gg.lake_nutes.length | Number of lake tiles on the board  |
        _total_nutrition = d["total_nutrition"]  # gg.lake_nutes.reduce((a,b) => a + b, 0) | Total nutrition of the lake
        #helpers
        # set class variables
        # set features

    def _extractFromSalestart(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _farmbit = d["farmbit"]  # farmbit_data_short(f) | Farmbit information reduced to an array.
        _item = d["item"]  # item_data_short(it)| Item information reduced to an array. See [Data Short](#DataShort)  |
        _worth = d["worth"]  # worth | monetary value of the item to be sold |
        #helpers
        # set class variables
        # set features

        item_type_str = self._ENUM_TO_STR["ITEM TYPE"][_item[2]]
        # items_on_screen[item_type_str] -= 1

        self.feature_time_since_start(f"time_to_first_{item_type_str}_sale")

    def _extractFromSaleend(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _farmbit = d["farmbit"]  # farmbit_data_short(f) | Farmbit information reduced to an array.
        _item = d["item"]  # item_data_short(it)| Item information reduced to an array. See [Data Short](#DataShort)  |
        _worth = d["worth"]  # worth | monetary value of the item sold |

        #helpers
        # set class variables
        # set features
        self.feature_inc("money_earned", _worth)

    def _extractFromRainstarted(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        # The log itself indicates that rain has started.

        #helpers
        # set class variables
        # set features

    def _extractFromEatfood(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
        _farmbit = d["farmbit"]  # farmbit_data_short(f) | Farmbit information reduced to an array.
        _item = d["item"]  # item_data_short(it)| Item information reduced to an array. See [Data Short](#DataShort)  |

        #helpers
        # set class variables
        # set features

    def _extractFromReset(self, event_client_time, event_data):
        # assign event_data variables
        d = event_data
         # The log itself indicates that the game has recieved the signal to reset.

        #helpers
        # set class variables
        # set features
        self.feature_time_since_start(feature_base=f"time_to_reset")

    def finish_window(self, w):
        window_end_time = timedelta(seconds=self._WINDOW_RANGES[w][1])
        for f in self.features._perlevel_names:
            cur_val = self.getValByIndex(f, w)
            if cur_val == float('inf'):
                # if min_ feature is still float('inf'), then max_value and min_value should still be 0.
                self.setValByIndex(f, w, new_value=0)
        if self._nutrition_view_on:
            self.set_time_in_nutrition_view(self._CLIENT_START_TIME + window_end_time)
            self._last_nutrition_open = self._CLIENT_START_TIME + window_end_time

        # update time_in_secs to full (might not be full

    def start_window(self, w):
        for f in self.features._perlevel_names:
            self.setValByIndex(f, w, new_value=self._get_default_val(f))
        if w > 0:
            for type in self._BUILDING_TYPES:
                # these types are immutable, so they will be at minimum the max previous value
                prev_val = self.getValByIndex(f'max_num_{type}_in_play', w-1)
                self.setValByIndex(f'min_num_{type}_in_play', w, prev_val)
                self.setValByIndex(f'max_num_{type}_in_play', w, prev_val)

    def _calculateAggregateFeatures(self):
        for w in self._cur_windows:
            self.finish_window(w)

    def set_time_in_nutrition_view(self, event_client_time):
        if not self._nutrition_view_on:
            self.feature_inc("time_in_nutrition_view", event_client_time - self._last_nutrition_open)

    def update_time_at_speed(self, event_client_time, prev_speed):
        # helpers
        prev_speed_str = LakelandExtractor._ENUM_TO_STR["SPEED"][prev_speed]
        time_at_prev_speed = event_client_time - self._last_speed_change
        
        # set class variables
        self._last_speed_change = event_client_time
        
        # set features
        self.feature_inc(f'time_in_game_speed_{prev_speed_str}', time_at_prev_speed)

    # Feature Type Setters 
    def feature_average(self, fname_base, value):
        if value == None:
            return
        win_pref, sess_pref = LakelandExtractor._WINDOW_PREFIX, LakelandExtractor._SESS_PREFIX
        window_feature, session_feature = win_pref + fname_base, sess_pref + fname_base
        for w in self._cur_windows:
            self.average_handler_window[window_feature][w]['total'] += value
            self.average_handler_window[window_feature][w]['n'] += 1
            avg = self.average_handler_window[window_feature][w]['total'] / \
                  self.average_handler_window[window_feature][w]['n']
            self.setValByIndex(window_feature, w, avg)

        self.average_handler_session[session_feature]['total'] += value
        self.average_handler_session[session_feature]['n'] += 1
        avg = self.average_handler_session[session_feature]['total'] / self.average_handler_session[session_feature][
            'n']
        self.setValByName(session_feature, avg)

    def feature_count(self, feature_base, windows=None):
        self.feature_inc(feature_base=feature_base, increment=1, windows=windows)

    def feature_inc(self, feature_base, increment, windows=None):
        win_pref, sess_pref = LakelandExtractor._WINDOW_PREFIX, LakelandExtractor._SESS_PREFIX
        self._increment_feature_in_cur_windows(feature_name=win_pref + feature_base, increment=increment, windows=windows)
        self._increment_sess_feature(feature_name=sess_pref + feature_base, increment=increment)

    def feature_time_since_start(self, feature_base):
        """
        Sets a session time since start feature. Will not write over a feature that has already been set.
        :param feature_base: name of feature without sess or window prefix
        :param cur_client_time: client time at which the event happened
        """
        feature_name = LakelandExtractor._SESS_PREFIX + feature_base
        feature_name_active = feature_name.replace('time_','time_active_')
        if self.getValByName(feature_name) in LakelandExtractor._NULL_FEATURE_VALS:
            self.setValByName(feature_name=feature_name, new_value=self.time_since_start(self.event_client_time))

        if self.getValByName(feature_name_active) in LakelandExtractor._NULL_FEATURE_VALS:
            time_active = self.getValByName('sess_time_active') or timedelta(0)
            self.setValByName(feature_name=feature_name_active, new_value=time_active)

    def feature_max_min(self, fname_base, val):
        """feature must have the same fname_base that gets put on the following four features:
        - {_SESS_PREFIX}max_
        - {_SESS_PREFIX}min_
        - {_WINDOW_PREFIX}max_
        - {_WINDOW_PREFIX}min_ """

        win_pref, sess_pref = LakelandExtractor._WINDOW_PREFIX, LakelandExtractor._SESS_PREFIX
        self._set_feature_max_in_cur_windows(feature_name=win_pref + "max_" + fname_base, val=val)
        self._set_feature_min_in_cur_windows(feature_name=win_pref + "min_" + fname_base, val=val)
        self._set_feature_max_in_session(feature_name=sess_pref + "max_" + fname_base, val=val)
        self._set_feature_min_in_session(feature_name=sess_pref + "min_" + fname_base, val=val)

    def _increment_feature_in_cur_windows(self, feature_name, increment=None, windows = None):
        if windows == None:
            windows = self._cur_windows
        if increment is None:
            increment = 1
        for w in windows:
            if self.getValByIndex(feature_name=feature_name, index=w) in LakelandExtractor._NULL_FEATURE_VALS:
                self.setValByIndex(feature_name, index=w, new_value=self._get_default_val(feature_name))
            self.features.incValByIndex(feature_name=feature_name, index=w, increment=increment)

    def _increment_sess_feature(self, feature_name, increment=None):
        if increment is None:
            increment = 1
        if self.getValByName(feature_name) in LakelandExtractor._NULL_FEATURE_VALS:
            self.setValByName(feature_name, new_value=self._get_default_val(feature_name))
        self.features.incAggregateVal(feature_name=feature_name, increment=increment)

    def _set_value_in_cur_windows(self, feature_name, value, windows = None):
        if windows == None:
            windows = self._cur_windows
        for w in windows:
            self.setValByIndex(feature_name=feature_name, index=w, new_value=value)

    def _set_feature_max_in_cur_windows(self, feature_name, val, windows = None):
        if windows == None:
            windows = self._cur_windows
        for w in windows:
            prev_val = self.getValByIndex(feature_name=feature_name, index=w)
            if val > prev_val:
                self.setValByIndex(feature_name=feature_name, index=w, new_value=val)

    def _set_feature_min_in_cur_windows(self, feature_name, val, windows = None):
        if windows == None:
            windows = self._cur_windows
        for w in windows:
            prev_val = self.getValByIndex(feature_name=feature_name, index=w)
            if val < prev_val:
                self.setValByIndex(feature_name=feature_name, index=w, new_value=val)

    def _set_feature_max_in_session(self, feature_name, val):
        prev_val = self.getValByName(feature_name=feature_name)
        if prev_val in LakelandExtractor._NULL_FEATURE_VALS or val > prev_val:
            self.setValByName(feature_name=feature_name, new_value=val)

    def _set_feature_min_in_session(self, feature_name, val):
        prev_val = self.getValByName(feature_name=feature_name)
        if prev_val in LakelandExtractor._NULL_FEATURE_VALS or val < prev_val:
            self.setValByName(feature_name=feature_name, new_value=val)

    def _get_default_val(self, feature_name):
        startswith = lambda prefix: feature_name.startswith(LakelandExtractor._SESS_PREFIX+prefix) or \
            feature_name.startswith(LakelandExtractor._WINDOW_PREFIX+prefix)
        if startswith('min_'):
            return float('inf')
        if startswith('time_'):
            return timedelta(0)
        else:
            return 0

    def getValByName(self, feature_name):
        return self.features.getValByName(feature_name)

    def setValByName(self, feature_name, new_value):
        self.features.setValByName(feature_name, new_value)

    def getValByIndex(self, feature_name, index):
        return self.features.getValByIndex(feature_name, index)

    def setValByIndex(self, feature_name, index, new_value):
        self.features.setValByIndex(feature_name, index, new_value)

    # Single Functions

    # unused
    # def _is_tutorial_mode_over(self):
    #     for tut, [tut_start, tut_end] in self._tutorial_start_and_end_times.items():
    #         if tut in LakelandExtractor._TUTORIAL_MODE and not tut_end:
    #             return False
    #     return True

    # def _avg_time_per_tutorial_mode_blurb(self):
    #     all_deltas = []
    #     for tut in LakelandExtractor._TUTORIAL_MODE:
    #         all_deltas.extend(self._tutorial_times_per_blurb[tut])
    #     return sum(all_deltas) / len(all_deltas)

    def _add_event(self, event_char):
        old = self.getValByName("event_sequence") or ''
        self.setValByName("event_sequence", old+str(event_char))

    def _change_population(self, client_time, increment):
        self._num_farmbits += increment
        self._max_population = max([self._num_farmbits, self._max_population])
        self._population_history.append((client_time, self._num_farmbits))

    def _get_pops_at_times(self, datetime_list):
        # takes in a sorted list of times and returns the population of farmbits at those times
        ret = []
        pop_i = 0
        for t in datetime_list:
            while not (pop_i + 1 >= len(self._population_history)) and self._population_history[pop_i + 1][0] < t:
                pop_i += 1
            ret.append(self._population_history[pop_i][1])
        return ret

    def _get_windows_at_time(self, client_time=None):
        # if no client time, get windows at start
        if client_time is not None:
            seconds_since_start = self.time_since_start(client_time).total_seconds()
        else:
            seconds_since_start = 0
        windows = []
        for i, (window_start, window_end) in self._WINDOW_RANGES.items():
            if window_start <= seconds_since_start < window_end:
                windows.append(i)

        return windows

    def _get_window_start_end_time(self, w):
        start_time = self._CLIENT_START_TIME + timedelta(
            seconds=self._WINDOW_RANGES[w][0])
        end_time = self._CLIENT_START_TIME + timedelta(
            seconds=self._WINDOW_RANGES[w][1])

        return (start_time, end_time)

    def _get_window_ranges(self):
        #window ranges are inclusive of the first number and exclusive of the second, i.e. (0, 30) is the range from
        # 0s to 29s
        window_start = 0
        window_end = 0
        window_ranges = {}
        for i in self.WINDOW_RANGE:
            window_end = window_end + self._NUM_SECONDS_PER_WINDOW
            window_ranges[i] = (window_start, window_end)
            window_start = window_end - self._NUM_SECONDS_PER_WINDOW_OVERLAP
        return window_ranges

    def _add_emote_to_window_by_time(self, emote):
        event_time = emote[-1]
        windows = self._get_windows_at_time(event_time)
        for window in windows:
            self._emotes_by_window[window].append(emote)

    def milliseconds_to_datetime(self, milliseconds):
        ret = datetime.utcfromtimestamp(milliseconds / 1000)
        assert ret >= self._CLIENT_START_TIME - timedelta(seconds=1)
        return ret

    def reformat_history_array(self, array, timestamp):
        """
        This function should be done in the reformatting step. History arrays contain timestamps relative to another
        timestamp that is when the array is sent through the logger. This function returns an array with absolute
        datetimes.
        :param self: self
        :param array: array to be processed. These arrays are either a list of times relative to client time or a list
        of subarrays with the last item as the time before timestamp.
        :param timestamp: time that each element is relative to (Javascript Date.now(), milliseconds since beginning 1970 UTC)
        :return: A 2D array where each subarray's final element is the absolute client time.
        """
        if not array:
            return []
        if type(array[0]) == int:
            return [self.milliseconds_to_datetime(t + timestamp) for t in array]
        else:
            ret = deepcopy(array)
            for i in range(len(ret)):
                ret[i][-1] = self.milliseconds_to_datetime(array[i][-1] + timestamp)
            return ret

    def get_tile_types_xys(self, type_list):
        if not self._TILE_OG_TYPES:
            self.log_warning("THERE ARE NO OG TYPES")
        else:
            return [LakelandExtractor.index_to_xy(i) for i, type in enumerate(self._TILE_OG_TYPES) if
                LakelandExtractor._ENUM_TO_STR['TILE TYPE'][type] in type_list]  # 5 is the lake type enum

    def time_since_start(self, client_time):
        return client_time - self._CLIENT_START_TIME

    def get_building_buildable_function(self):
        buildable_area_around_building = lambda bx, by:set((x,y) for x in range(bx-2, bx+3) for y in range(by-2, by+3))
        buildable_xys = set()
        for bx, by in self.building_xys:
            buildable_xys = buildable_xys | buildable_area_around_building(bx, by)

        # does not consider the small area in the beginning by the lake that players get for free.
        def building_buildable(t):
            type, tx, ty = t[3], t[4], t[5]
            if type != 1: # land
                return False
            if not self.in_bounds(tx, ty):
                return False
            return (tx,ty) in buildable_xys

        return building_buildable

    def in_bounds(self, tx, ty):
        # initial bounds 17-30 for 0 or 1 max farmbits.
        min_bound, max_bound = 17, 30
        # intial bounds 16-31 for 2 max farmbits, etc.
        if self._max_population > 1:
            change = self._max_population - 1
            min_bound = max(0, min_bound - self._max_population)
            max_bound = min(50, max_bound - self._max_population)
        return all(min_bound <= z <= max_bound for z in [tx, ty])

    def add_debug_str(self, s):
        self.debug_strs.append(s)
        
    def log_warning(self, message):
        self.add_debug_str('WARNING: '+message)
        debug_str = '\n'.join(self.debug_strs)
        Logger.Log(debug_str, logging.WARNING)
        self.debug_strs = []
        
    def reset(self):
        self.levels:           List[int]       = []
        self.last_adjust_type: Optional[str] = None
        self.features:         LegacyFeature.LegacySessionFeatures = LegacyFeature.LegacySessionFeatures(game_schema=self._GAME_SCHEMA)
        self.setValByName('sessID', new_value=self._session_id)

        # Initialize Lakeland Variables
        # Constants
        self._TILE_OG_TYPES = []
        self._CLIENT_START_TIME = None
        self._VERSION = "0"
        

        # Feature Setting Vars
        self._cur_windows = self._get_windows_at_time()
        self.average_handler_window = defaultdict(lambda: {k: {'n': 0, 'total': 0} for k in self.WINDOW_RANGE})
        self.average_handler_session = defaultdict(lambda: {'n': 0, 'total': 0})

        # Flags
        self._active = True
        self._nutrition_view_on = False
        self._first_gamestate = True
        self._debug = False
        self._continue = None
        self._end = False

        # Timestamps
        self._last_active_event_time = None

        self._last_speed_change = None
        self._last_nutrition_open = None
        self._last_event_time = None

        # Gamestate Trackers: Continue Independent
        self._selected_buy_cost = 0
        self._emotes_by_window = defaultdict(lambda: [])
        self._cur_speed = 2  # default is play, 2
        # TODO: Check and see if all the tutorials are correct with skips/continues etc
        self._tutorial_times_per_blurb = {key: [] for key in LakelandExtractor._STR_TO_ENUM['TUTORIALS']}
        # self._time_per_speed = {key: 0 for key in LakelandExtractor._STR_TO_ENUM['SPEED']}
        # self._tutorial_start_and_end_times = {key: [None, None] for key in LakelandExtractor._STR_TO_ENUM['TUTORIALS']}
        # self._tutorial_compliance = {tut: None for tut in LakelandExtractor._TUTORIAL_MODE}

        # Gamestate Trackers: Continue Dependent
        self._num_farmbits = 0
        self._population_history = [(datetime.min, 0)]  # client time, population
        self.building_xys = []
        self._tiles = defaultdict(lambda: {'marks': [1, 1, 1, 1], 'type': None}) #maps (tx,ty) => {marks: [], type: int}
        self._max_population = 0

        # start first windows:
        for w in self._cur_windows:
            self.start_window(w)

    @staticmethod
    def onscreen_item_dict():
        ret = {key: 0 for key in LakelandExtractor._BUILDING_TYPES + ['food', 'poop', 'fertilizer', 'milk']}
        return ret

    @staticmethod
    def index_to_xy(index):
        x = index % 50
        y = index // 50
        return x, y

def avg(l):
    """
    Returns average of a list. Allows elements to be datetime time intervals.
    :param l: list
    :return: Average (L1 distance between elements)
    """
    if len(l) == 0:
        return None
    if len(l) == 1:
        return l[0]

    lsum = l[0]
    for d in l[1:]:
        lsum += d
    return lsum / len(l)

def list_deltas(l):
    return [l[i] - l[i - 1] for i in range(1, len(l))]

def distance(point1, point2):
    """

    :param point1: x,y for point 1
    :param point2: x,y for point 2
    :return: euclidean distance between x and y
    """
    x1, y1 = point1[0], point1[1]
    x2, y2 = point2[0], point2[1]
    return sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

def get_tile_txy(tile):
    return tile[4], tile[5]

def tile_i_to_xy(i):
    ty = i//50
    tx = i%50
    return tx,ty



# import standard libraries
import logging
from typing import Any, Dict, Optional
# import local files
from ogd.core.schemas.games.DataElementSchema import DataElementSchema
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class EventSchema(Schema):
    """
    Dumb struct to contain a specification of an Event in a GameSchema file.

    These essentially are just a description of the event, and a set of elements in the EventData attribute of the Event.
    """
    def __init__(self, name:str, all_elements:Dict[str, Dict]):
        self._description : str                               = "No description available"
        self._event_data  : Dict[str, DataElementSchema] = {}

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} Event config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "description" in all_elements.keys():
            self._description = EventSchema._parseDescription(description=all_elements['description'])
        else:
            self._description = "Unknown"
            Logger.Log(f"{name} EventSchema config does not have a 'description' element; defaulting to description='{self._description}", logging.WARN)
        if "event_data" in all_elements.keys():
            self._event_data = EventSchema._parseEventDataElements(event_data=all_elements['event_data'])
        else:
            self._event_data = {}
            Logger.Log(f"{name} EventSchema config does not have an 'event_data' element; defaulting to empty dict", logging.WARN)
        _leftovers = { key : val for key,val in all_elements.items() if key not in {"description", "event_data"} }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def Description(self) -> str:
        return self._description

    @property
    def EventData(self) -> Dict[str, DataElementSchema]:
        return self._event_data

    @property
    def AsMarkdown(self) -> str:
        return "\n\n".join([
            f"### **{self.Name}**",
            self.Description,
            "#### Event Data",
            "\n".join(
                  [elem.AsMarkdown for elem in self.EventData.values()]
                + (["- Other Elements:"] +
                   [f"  - **{elem_name}**: {elem_desc}" for elem_name,elem_desc in self.NonStandardElements]
                  ) if len(self.NonStandardElements) > 0 else []
            )
        ])

    @property
    def AsMarkdownTable(self) -> str:
        return "\n\n".join([
            f"### **{self.Name}**",
            f"{self.Description}",
            "#### Event Data",
            "\n".join(
                ["| **Name** | **Type** | **Description** | **Sub-Elements** |",
                 "| ---      | ---      | ---             | ---         |"]
              + [elem.AsMarkdownRow for elem in self.EventData.values()]
            ),
            "#### Other Elements",
            "\n".join(
                [f"- **{elem_name}**: {elem_desc}  " for elem_name,elem_desc in self.NonStandardElements]
                if len(self.NonStandardElements) > 0 else ["- None"]
            )
        ])

    @staticmethod
    def _parseEventDataElements(event_data):
        ret_val : Dict[str, DataElementSchema]
        if isinstance(event_data, dict):
            ret_val = {name:DataElementSchema(name=name, all_elements=elems) for name,elems in event_data.items()}
        else:
            ret_val = {}
            Logger.Log(f"event_data was unexpected type {type(event_data)}, defaulting to empty dict.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDescription(description):
        ret_val : str
        if isinstance(description, str):
            ret_val = description
        else:
            ret_val = str(description)
            Logger.Log(f"Event description was not a string, defaulting to str(description) == {ret_val}", logging.WARN)
        return ret_val



# import libraries
from typing import Any, Callable, Dict, List, Type, Optional
# import locals
from ogd.core.generators.registries.DetectorRegistry import DetectorRegistry
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.processors.GeneratorProcessor import GeneratorProcessor
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.utils import ExportRow

class DetectorProcessor(GeneratorProcessor):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_schema: GameSchema, LoaderClass: Type[GeneratorLoader], trigger_callback:Callable[[Event], None],
                 feature_overrides:Optional[List[str]]=None):
        # TODO: Consider having multiple registries for per-player or per-session kinds of things.
        super().__init__(game_schema=game_schema, LoaderClass=LoaderClass, feature_overrides=feature_overrides)
        self._registry = DetectorRegistry(mode=self._mode, trigger_callback=trigger_callback)
        self._registry.LoadFromSchema(schema=game_schema, loader=self._loader, overrides=feature_overrides)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @property
    def _mode(self) -> ExtractionMode:
        return ExtractionMode.DETECTOR

    @property
    def _playerID(self) -> str:
        return "detectors"

    @property
    def _sessionID(self) -> str:
        return "detectors"

    def _getExtractorNames(self, order:int) -> Dict[str,List[FeatureData]]:
        raise NotImplementedError("Function stub! Haven't written name getter for detector processor.")

    def _processEvent(self, event:Event):
        if self._registry is not None:
            self._registry.UpdateFromEvent(event)

    def _getLines(self) -> List[ExportRow]:
        return []

    def _clearLines(self):
        if self._registry is not None:
            self._registry.LoadFromSchema(schema=self._game_schema, loader=self._loader, overrides=self._overrides)

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***
